#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging,os,socket,select, sys,threading, time

import scapy.config

from lrucache import LRUCache
from ctypes_openssh import AES_BLOCK_SIZE

log=logging.getLogger('socket.scapy')

WAIT_RETRANSMIT=5
MISSING_DATA_MESSAGE='[MISSINGDATA456]'

def isdestport22(packet):
  return  packet.dport == 22

def isNotdestport22(packet):
  return  not isdestport22(packet)


def hexify(data):
  s=''
  for i in range(0,len(data)):
    s+="%02x"% ord(data[i])
    if i%16==15:
      s+="\r\n"
    elif i%2==1:
      s+=" "
  #s+="\r\n"
  return s


class MissingDataException (Exception):
  def __init__(self,nb):
    self.nb=nb
 
class state:
  name=None
  packet_count=0
  byte_count=0
  start_seq=None
  max_seq=0
  expected_seq=0
  queue=None
  write_socket=None
  read_socket=None
  read_socket_revc_func=None
  ts_missing=None
  def __init__(self,name):
    self.name=name
    self.queue={}
  def init(self, pair):
    read, write = pair
    self.read_socket=socket.socket(_sock=read)
    self.read_socket_revc_func=self.read_socket.recv
    #self.read_socket.recv = self._wrapped_recv # will be used to lock-wrap
    self.write_socket=write
    return  

  def enqueue(self, packet):
    self.queue[packet.seq]= packet
    return
    

  def forget_missing_data(self):
    ''' switch recv() method on readsocket to alert reader()
     push missing data in socket to wakeup reader
    '''
    if self.ts_missing is None: # naaa. it's good
      return
    self.ts_missing=None
    #
    if self.read_socket.recv != self._read_missing_data:
      log.warning('socket empty... we switch to Missing mode => send(MISSING_DATA_MESSAGE)')
      self.read_socket.recv = self._read_missing_data
    return
    
  def _read_missing_data(self, n):
    ''' exception contains lenght of missing data '''
    # repair socket for further use
    self.read_socket.recv = self.read_socket_revc_func
    #self.read_socket.recv = self._wrapped_recv    
    nb = self.expected_seq - self.max_seq
    new_seq = min(self.queue.keys())
    nbreal = new_seq - self.max_seq
    log.debug('%s: Forgetting about r:%d e:%d bytes / %d pkts left, expected: %d, new_seq: %d'%(
                            self.name, nbreal, nb, len(self.queue), self.expected_seq ,new_seq ))
    raise MissingDataException(nbreal)
  
  def __str__(self):
    return "%s: %d bytes/%d packets max_seq:%d expected_seq:%d q:%d"%(self.name, self.byte_count,self.packet_count,
                self.max_seq,self.expected_seq, len(self.queue))
  
class stack:
  def __init__(self):
    self.inbound=state('inbound')
    self.outbound=state('outbound')
  def __str__(self):
    return "\n%s\n%s"%(self.inbound,self.outbound)
  
class socket_scapy():
  ''' what you write in writeso, gets read in readso '''
  worker=None
  def __init__(self,filterRules, protocolName='TCP', packetCount=0, timeout=None,
            isInboundPacketCallback=None,isOutboundPacketCallback=None):
    ''' 
    @param filterRules: a pcap compatible filter string 
    @param protocolName: the name of the scapy proto layer 
    @param packetCount: 0/Unlimited or packet capture limit
    @param timeout: None/Unlimited or stop after
    '''
    # set scapy to use native pcap instead of SOCK_RAW
    scapy.config.conf.use_pcap=True

    self._cache_seqn = LRUCache(128)
    ## if using SOCK_RAW, we need to mess with filter to up the capture size higher than 1514/1600 bytes
    #maxSize="\' -s \'0xffff" # abusing scapy-to-tcpdump string format 
    #self.filterRules=filterRules + maxSize
    self.filterRules=filterRules
    self.protocolName=protocolName
    self.packetCount=packetCount
    self.timeout=timeout
    #
    self.lock=threading.Lock()
    self.stack=stack()
    self.packets={}
    self.bigqueue=[]

    self._running_thread=None
    # distinguish between incoming and outgoing packets // classic ssh client
    self.__is_inboundPacket=isInboundPacketCallback
    self.__is_outboundPacket=isOutboundPacketCallback
    if ( self.__is_inboundPacket is None):
      self.__is_inboundPacket=isNotdestport22 ## SSH CLIENT
      ##self.__is_inboundPacket=isdestport22  ## SSH SERVER
    if ( self.__is_outboundPacket is None):
      self.__is_outboundPacket=isdestport22  ## SSH CLIENT
      ##self.__is_outboundPacket=isNotdestport22 ## SSH SERVER
    # make socket
    try:
        isWindows = socket.AF_UNIX
        self.stack.inbound.init(  socket.socketpair() )
        self.stack.outbound.init( socket.socketpair() )
    except NameError:
        # yes || no socketpair support anyway
        self._initPipes()
    # scapy config
    # loopback
    #log.info('MTU is %d'%(scapy.data.MTU ))
    #scapy.data.MTU=0x7fff
    return
  
  def _initPipes(self):
    self.stack.inbound.pipe  = pipe_socketpair()
    self.stack.inbound.init(  self.stack.inbound.pipe.socketpair())
    self.stack.outbound.pipe = pipe_socketpair()
    self.stack.outbound.init( self.stack.outbound.pipe.socketpair() )
    return

  def getInboundSocket(self):
    return self.stack.inbound.read_socket
  def getOutboundSocket(self):
    return self.stack.outbound.read_socket

  def run(self):
    # scapy - with config initialised
    #scapy.sendrecv.sniff(count=self.packetCount,timeout=self.timeout,store=0,filter=self.filterRules,prn=self.cbSSHPacket)
    from scapy.all import sniff
    log.info('Using L2listen = %s'%(scapy.config.conf.L2listen)) 
    # XXX TODO, define iface from saddr and daddr // scapy.all.read_routes()
    sniff(count=self.packetCount, timeout=self.timeout, store=0, filter=self.filterRules, prn=self.enqueue, iface='any')
    log.warning('============ SNIFF Terminated ====================')
    return

  def enqueue(self, obj):
    self.lock.acquire()
    self.bigqueue.append(obj)
    self.lock.release()
    return
  def dequeue(self):
    obj=None
    self.lock.acquire()
    if len(self.bigqueue) > 0:
      obj=self.bigqueue.pop(0)
    self.lock.release()
    return obj

  def prequeue(self, state):
    ''' get packets from queue and put them in processing bigqueue '''
    queue=state.queue.values()
    queue.sort(key=lambda p: p.seq)
    # add the first and the next that are in perfect suite.
    toadd=[queue.pop(0),]
    for i in xrange(0,len(queue)):
      if toadd[-1].seq+len(toadd[-1].payload) == queue[0].seq :
        toadd.append(queue.pop(0))
      else:
        log.debug('Stop prequeuing  toadd[-1].seq+len(toadd[-1].payload): %d , queue[0].seq: %d'%(toadd[-1].seq+len(toadd[-1].payload) , queue[0].seq ))
        break
    # let remaining in state.queue
    state.queue=dict( [ (p.seq, p) for p in queue])
    self.lock.acquire()
    #preprend packets
    self.bigqueue = toadd + self.bigqueue
    self.lock.release()
    log.debug('Prequeued %d packets, remaining %d, queued from %d to %d '%(
                        len(toadd), len(state.queue), toadd[0].seq, (toadd[-1].seq + len(toadd[-1].payload)) 
                        ))
    # reset time counter
    if len(state.queue) > 0:
      state.ts_missing  = time.time()
    else:
      state.ts_missing  = None
    return
  
  def run2(self):
    while(True):
      # depile packets
      packet=self.dequeue()
      self.cbSSHPacket(packet)
      if packet is None:
        if time.time() %10 == 0:
          log.debug('%s'%(self))
        time.sleep(0.5)
    log.warning('============ SNIFF WORKER Terminated ====================')
    return
    
  def checkState(self,state, packet):
    seq=packet.seq    
    ## debug head initialisation
    if state.start_seq is None:
      state.start_seq=seq
      state.expected_seq=seq
    ## we should not processs empty packets
    payloadLen=len(packet.payload)
    if payloadLen == 0:
      return False

    # packet is expected 
    if seq == state.expected_seq: # JIT
      state.max_seq = seq
      state.expected_seq = seq + payloadLen
      self.packets[seq]=packet # debug the packets
      state.ts_missing = None
      # check if next is already in state.queue , if expected has changed
      if state.expected_seq in state.queue : 
        log.debug('in state.queue: seqnum %d len: %d %s'%(seq, payloadLen, state ))
        log.debug('we have next ones in buffer %d'%(state.expected_seq)) # prepend the queue to parse list.
        self.prequeue(state)
      return True

    # packet is future
    elif seq > state.expected_seq: # future anterieur
      # seq is in advance
      state.enqueue(packet)
      #log.debug('Queuing packet seq: %d len: %d %s'%(seq, payloadLen, state))
      ## if ~5 sec, we will forget about missing data
      ## and inform reader of that fact
      if state.ts_missing is None:
        state.ts_missing = time.time()
      elif time.time() > ( state.ts_missing + WAIT_RETRANSMIT) : # waiting for too long
        state.forget_missing_data()
      else:
        pass
      # check if next is already in state.queue 
      if state.expected_seq in state.queue :
        log.debug('queuing, seqnum %d len: %d %s'%(seq, payloadLen, state ))
        log.debug('queuing, we have next ones in buffer %d'%(state.expected_seq)) # prepend the queue to parse list.
        self.prequeue(state)
        # reset time counter
        state.ts_missing  = None
      else:
        ## if ~5 sec, we will forget about missing data
        ## and inform reader of that fact
        if state.ts_missing is None:
          state.ts_missing = time.time()
        elif time.time() > ( state.ts_missing + WAIT_RETRANSMIT) : # waiting for too long
          state.forget_missing_data()
        else:
          pass
      #do not parse this one 
      return False

    # packet is a retranmission 
    elif seq < state.expected_seq:
      # TCP retransmission
      log.debug('TCP retransmit - We just received %d when we already processed %d'%(seq, state.max_seq))
      # never hitted
      if seq+payloadLen > state.expected_seq :
        log.warning(' ***** EXTRA DATA FOUND ON TCP RETRANSMISSION ')
        # we need to recover the extra data to put it in the stream
        nb=(seq+payloadLen) - state.expected_seq
        data=packet.payload.load[-nb:]
        log.warning('packet seq %d has already been received'%(seq))
        log.warning('recent  packet : %s'%(repr(packet.underlayer) ))
        log.warning('first   packet : %s'%(repr(self.packets[seq].underlayer ) ))
        seq2=seq+len(self.packets[seq].payload)
        log.warning('first+1 packet : %s'%(repr(self.packets[seq2].underlayer ) ))
        packet.payload.load=data
        log.warning('NEW     packet : %s'%(repr(packet.underlayer) ))
        # updates seq
        state.max_seq = seq
        state.expected_seq = seq2
        # save it
        self.packets[seq]=packet # debug the packets
        # we can process this one
        return True
      # ignore it
      return False
    # else, seq < expected_seq and seq >= state.max_seq. That's not possible... it's current packet.
    # it's a dup already dedupped ? partial fragment ?
    log.warning('received a bogus fragment seq: %d for %s'%(seq, self))
    return True

  def cbSSHPacket(self, obj):
    ''' callback function to pile packets in socket'''
    # check queues only
    if obj is None:
      for state in [self.stack.inbound,self.stack.outbound]:
        if state.expected_seq in state.queue : 
          self.prequeue(state)
        if state.ts_missing is not None and  time.time() > ( state.ts_missing + WAIT_RETRANSMIT) : # waiting for too long
            state.forget_missing_data()
      return None
    # triage
    packet=obj[self.protocolName]
    pLen=len(packet.payload)
    if pLen == 0: # ignore acks and stuff 
      return None
    # real triage
    if self.__is_inboundPacket(packet):
      if self.checkState(self.stack.inbound, packet) and pLen > 0:
        self.writePacket(self.stack.inbound, packet.payload.load )
    elif self.__is_outboundPacket(packet):
      if self.checkState(self.stack.outbound, packet) and pLen > 0:
        self.writePacket(self.stack.outbound, packet.payload.load )
    else:
      log.error('the packet is neither inbound nor outbound. You messed up your filter and callbacks.')
    return None
    
  def setThread(self,thread):
    ''' the thread using the pipes '''
    self._running_thread=thread
    return 
  
  def writePacket(self,state, payload):
    #if state.name == 'inbound':
    #  log.debug("writePacket%s %d + len: %d = %d"%(state.name, state.byte_count, len(payload), state.byte_count+ len(payload) ) )

    state.byte_count+=self.addPacket(payload,state)
    state.packet_count+=1
    #log.debug("writePacket%s %d len: %d\n%s"%(state.name, state.byte_count, len(payload), hexify(payload) ) )
    return 
    
  def addPacket(self,payload, state):
    cnt=state.write_socket.send(payload)
    #log.debug("buffered %d/%d bytes"%(cnt, len(payload) ) )
    return cnt
  
  def __str__(self):
    return "<socket_scapy %s "%(self.stack) 

# if Linux use socket.socketpair()
class pipe_socketpair(object):
  def __init__(self):
    self.readfd,self.writefd=os.pipe()
    self.readso=socket.fromfd(self.readfd,socket.AF_UNIX,socket.SOCK_STREAM)
    self.writeso=socket.fromfd(self.writefd,socket.AF_UNIX,socket.SOCK_STREAM)
    return
  def socketpair(self):
    return (self.readso,self.writeso)

def test():
  '''
sniff(count=0, store=1, offline=None, prn=None, lfilter=None, L2socket=None, timeout=None, opened_socket=None, *arg, **karg)  
  '''
  port=22
  sshfilter="tcp and port %d"%(port)
  soscapy=socket_scapy(sshfilter,packetCount=10)
  log.info('Please make some ssh  traffic')
  soscapy.run()
  print 'sniff finished'
  # we get Ether()'s...
  print soscapy.stats()
  l=soscapy._inbound_cnt
  print 'trying to read'
  data=soscapy.getInboundSocket().recv(l)
  print 'recv %d bytes ->',len(data),repr(data)
  return soscapy

#test()




from ctypes import *

STRING = c_char_p


class asn1_string_st(Structure):
    pass
ASN1_INTEGER = asn1_string_st
ASN1_ENUMERATED = asn1_string_st
ASN1_BIT_STRING = asn1_string_st
ASN1_OCTET_STRING = asn1_string_st
ASN1_PRINTABLESTRING = asn1_string_st
ASN1_T61STRING = asn1_string_st
ASN1_IA5STRING = asn1_string_st
ASN1_GENERALSTRING = asn1_string_st
ASN1_UNIVERSALSTRING = asn1_string_st
ASN1_BMPSTRING = asn1_string_st
ASN1_UTCTIME = asn1_string_st
ASN1_TIME = asn1_string_st
ASN1_GENERALIZEDTIME = asn1_string_st
ASN1_VISIBLESTRING = asn1_string_st
ASN1_UTF8STRING = asn1_string_st
ASN1_BOOLEAN = c_int
ASN1_NULL = c_int
class bignum_st(Structure):
    pass
BIGNUM = bignum_st
class bignum_ctx(Structure):
    pass
BN_CTX = bignum_ctx
class bn_blinding_st(Structure):
    pass
BN_BLINDING = bn_blinding_st
class bn_mont_ctx_st(Structure):
    pass
BN_MONT_CTX = bn_mont_ctx_st
class bn_recp_ctx_st(Structure):
    pass
BN_RECP_CTX = bn_recp_ctx_st
class bn_gencb_st(Structure):
    pass
BN_GENCB = bn_gencb_st
class buf_mem_st(Structure):
    pass
BUF_MEM = buf_mem_st
class evp_cipher_st(Structure):
    pass
EVP_CIPHER = evp_cipher_st
class evp_cipher_ctx_st(Structure):
    pass
EVP_CIPHER_CTX = evp_cipher_ctx_st
class env_md_st(Structure):
    pass
EVP_MD = env_md_st
class env_md_ctx_st(Structure):
    pass
EVP_MD_CTX = env_md_ctx_st
class evp_pkey_st(Structure):
    pass
EVP_PKEY = evp_pkey_st
class dh_st(Structure):
    pass
DH = dh_st
class dh_method(Structure):
    pass
DH_METHOD = dh_method
class dsa_st(Structure):
    pass
DSA = dsa_st
class dsa_method(Structure):
    pass
DSA_METHOD = dsa_method
class rsa_st(Structure):
    pass
RSA = rsa_st
class rsa_meth_st(Structure):
    pass
RSA_METHOD = rsa_meth_st
class rand_meth_st(Structure):
    pass
RAND_METHOD = rand_meth_st
class ecdh_method(Structure):
    pass
ECDH_METHOD = ecdh_method
class ecdsa_method(Structure):
    pass
ECDSA_METHOD = ecdsa_method
class x509_st(Structure):
    pass
X509 = x509_st
class X509_algor_st(Structure):
    pass
X509_ALGOR = X509_algor_st
class X509_crl_st(Structure):
    pass
X509_CRL = X509_crl_st
class X509_name_st(Structure):
    pass
X509_NAME = X509_name_st
class x509_store_st(Structure):
    pass
X509_STORE = x509_store_st
class x509_store_ctx_st(Structure):
    pass
X509_STORE_CTX = x509_store_ctx_st
class ssl_st(Structure):
    pass
SSL = ssl_st
class ssl_ctx_st(Structure):
    pass
SSL_CTX = ssl_ctx_st
class v3_ext_ctx(Structure):
    pass
X509V3_CTX = v3_ext_ctx
class conf_st(Structure):
    pass
CONF = conf_st
class store_st(Structure):
    pass
STORE = store_st
class store_method_st(Structure):
    pass
STORE_METHOD = store_method_st
class ui_st(Structure):
    pass
UI = ui_st
class ui_method_st(Structure):
    pass
UI_METHOD = ui_method_st
class st_ERR_FNS(Structure):
    pass
ERR_FNS = st_ERR_FNS
class engine_st(Structure):
    pass
ENGINE = engine_st
class X509_POLICY_NODE_st(Structure):
    pass
X509_POLICY_NODE = X509_POLICY_NODE_st
class X509_POLICY_LEVEL_st(Structure):
    pass
X509_POLICY_LEVEL = X509_POLICY_LEVEL_st
class X509_POLICY_TREE_st(Structure):
    pass
X509_POLICY_TREE = X509_POLICY_TREE_st
class X509_POLICY_CACHE_st(Structure):
    pass
X509_POLICY_CACHE = X509_POLICY_CACHE_st
class crypto_ex_data_st(Structure):
    pass
CRYPTO_EX_DATA = crypto_ex_data_st
CRYPTO_EX_new = CFUNCTYPE(c_int, c_void_p, c_void_p, POINTER(CRYPTO_EX_DATA), c_int, c_long, c_void_p)
CRYPTO_EX_free = CFUNCTYPE(None, c_void_p, c_void_p, POINTER(CRYPTO_EX_DATA), c_int, c_long, c_void_p)
CRYPTO_EX_dup = CFUNCTYPE(c_int, POINTER(CRYPTO_EX_DATA), POINTER(CRYPTO_EX_DATA), c_void_p, c_int, c_long, c_void_p)
class ocsp_req_ctx_st(Structure):
    pass
OCSP_REQ_CTX = ocsp_req_ctx_st
class ocsp_response_st(Structure):
    pass
OCSP_RESPONSE = ocsp_response_st
class ocsp_responder_id_st(Structure):
    pass
OCSP_RESPID = ocsp_responder_id_st
size_t = c_uint
__u_char = c_ubyte
__u_short = c_ushort
__u_int = c_uint
__u_long = c_ulong
__int8_t = c_byte
__uint8_t = c_ubyte
__int16_t = c_short
__uint16_t = c_ushort
__int32_t = c_int
__uint32_t = c_uint
__int64_t = c_longlong
__uint64_t = c_ulonglong
__quad_t = c_longlong
__u_quad_t = c_ulonglong
__dev_t = __u_quad_t
__uid_t = c_uint
__gid_t = c_uint
__ino_t = c_ulong
__ino64_t = __u_quad_t
__mode_t = c_uint
__nlink_t = c_uint
__off_t = c_long
__off64_t = __quad_t
__pid_t = c_int
__clock_t = c_long
__rlim_t = c_ulong
__rlim64_t = __u_quad_t
__id_t = c_uint
__time_t = c_long
__useconds_t = c_uint
__suseconds_t = c_long
__daddr_t = c_int
__swblk_t = c_long
__key_t = c_int
__clockid_t = c_int
__timer_t = c_void_p
__blksize_t = c_long
__blkcnt_t = c_long
__blkcnt64_t = __quad_t
__fsblkcnt_t = c_ulong
__fsblkcnt64_t = __u_quad_t
__fsfilcnt_t = c_ulong
__fsfilcnt64_t = __u_quad_t
__ssize_t = c_int
__loff_t = __off64_t
__qaddr_t = POINTER(__quad_t)
__caddr_t = STRING
__intptr_t = c_int
__socklen_t = c_uint
class _IO_FILE(Structure):
    pass
FILE = _IO_FILE
__FILE = _IO_FILE
_G_int16_t = c_short
_G_int32_t = c_int
_G_uint16_t = c_ushort
_G_uint32_t = c_uint
__gnuc_va_list = STRING
_IO_lock_t = None
__io_read_fn = CFUNCTYPE(__ssize_t, c_void_p, STRING, size_t)
__io_write_fn = CFUNCTYPE(__ssize_t, c_void_p, STRING, size_t)
__io_seek_fn = CFUNCTYPE(c_int, c_void_p, POINTER(__off64_t), c_int)
__io_close_fn = CFUNCTYPE(c_int, c_void_p)
cookie_read_function_t = __io_read_fn
cookie_write_function_t = __io_write_fn
cookie_seek_function_t = __io_seek_fn
cookie_close_function_t = __io_close_fn
class _IO_cookie_io_functions_t(Structure):
    pass
cookie_io_functions_t = _IO_cookie_io_functions_t
va_list = __gnuc_va_list
off_t = __off_t
off64_t = __off64_t
ssize_t = __ssize_t
class _G_fpos_t(Structure):
    pass
fpos_t = _G_fpos_t
class _G_fpos64_t(Structure):
    pass
fpos64_t = _G_fpos64_t
class __locale_struct(Structure):
    pass
__locale_t = POINTER(__locale_struct)
locale_t = __locale_t
u_char = __u_char
u_short = __u_short
u_int = __u_int
u_long = __u_long
quad_t = __quad_t
u_quad_t = __u_quad_t
class __fsid_t(Structure):
    pass
fsid_t = __fsid_t
loff_t = __loff_t
ino_t = __ino_t
ino64_t = __ino64_t
dev_t = __dev_t
gid_t = __gid_t
mode_t = __mode_t
nlink_t = __nlink_t
uid_t = __uid_t
pid_t = __pid_t
id_t = __id_t
daddr_t = __daddr_t
caddr_t = __caddr_t
key_t = __key_t
clock_t = __clock_t
time_t = __time_t
clockid_t = __clockid_t
timer_t = __timer_t
useconds_t = __useconds_t
suseconds_t = __suseconds_t
ulong = c_ulong
ushort = c_ushort
uint = c_uint
int8_t = c_int8
int16_t = c_int16
int32_t = c_int32
int64_t = c_int64
u_int8_t = c_ubyte
u_int16_t = c_ushort
u_int32_t = c_uint
u_int64_t = c_ulonglong
register_t = c_int
__sig_atomic_t = c_int
class __sigset_t(Structure):
    pass
sigset_t = __sigset_t
__fd_mask = c_long
fd_mask = __fd_mask
blksize_t = __blksize_t
blkcnt_t = __blkcnt_t
fsblkcnt_t = __fsblkcnt_t
fsfilcnt_t = __fsfilcnt_t
blkcnt64_t = __blkcnt64_t
fsblkcnt64_t = __fsblkcnt64_t
fsfilcnt64_t = __fsfilcnt64_t
pthread_t = c_ulong
class __pthread_internal_slist(Structure):
    pass
__pthread_slist_t = __pthread_internal_slist
pthread_key_t = c_uint
pthread_once_t = c_int
pthread_spinlock_t = c_int
__compar_fn_t = CFUNCTYPE(c_int, c_void_p, c_void_p)
comparison_fn_t = __compar_fn_t
__compar_d_fn_t = CFUNCTYPE(c_int, c_void_p, c_void_p, c_void_p)
class stack_st(Structure):
    pass
STACK = stack_st
class openssl_item_st(Structure):
    pass
OPENSSL_ITEM = openssl_item_st
class bio_st(Structure):
    pass
BIO_dummy = bio_st
class crypto_ex_data_func_st(Structure):
    pass
CRYPTO_EX_DATA_FUNCS = crypto_ex_data_func_st
class st_CRYPTO_EX_DATA_IMPL(Structure):
    pass
CRYPTO_EX_DATA_IMPL = st_CRYPTO_EX_DATA_IMPL
CRYPTO_MEM_LEAK_CB = CFUNCTYPE(c_void_p, c_ulong, STRING, c_int, c_int, c_void_p)
BIO = bio_st
bio_info_cb = CFUNCTYPE(None, POINTER(bio_st), c_int, STRING, c_int, c_long, c_long)
class bio_method_st(Structure):
    pass
BIO_METHOD = bio_method_st
class bio_f_buffer_ctx_struct(Structure):
    pass
BIO_F_BUFFER_CTX = bio_f_buffer_ctx_struct
class asn1_ctx_st(Structure):
    pass
ASN1_CTX = asn1_ctx_st
class asn1_const_ctx_st(Structure):
    pass
ASN1_const_CTX = asn1_const_ctx_st
class asn1_object_st(Structure):
    pass
ASN1_OBJECT = asn1_object_st
ASN1_STRING = asn1_string_st
class ASN1_ENCODING_st(Structure):
    pass
ASN1_ENCODING = ASN1_ENCODING_st
class asn1_string_table_st(Structure):
    pass
ASN1_STRING_TABLE = asn1_string_table_st
class ASN1_TEMPLATE_st(Structure):
    pass
ASN1_TEMPLATE = ASN1_TEMPLATE_st
class ASN1_ITEM_st(Structure):
    pass
ASN1_ITEM = ASN1_ITEM_st
class ASN1_TLC_st(Structure):
    pass
ASN1_TLC = ASN1_TLC_st
class ASN1_VALUE_st(Structure):
    pass
ASN1_VALUE = ASN1_VALUE_st
i2d_of_void = CFUNCTYPE(c_int, c_void_p, POINTER(POINTER(c_ubyte)))
d2i_of_void = CFUNCTYPE(c_void_p, POINTER(c_void_p), POINTER(POINTER(c_ubyte)), c_long)
ASN1_ITEM_EXP = ASN1_ITEM
class asn1_type_st(Structure):
    pass
ASN1_TYPE = asn1_type_st
class asn1_method_st(Structure):
    pass
ASN1_METHOD = asn1_method_st
class asn1_header_st(Structure):
    pass
ASN1_HEADER = asn1_header_st
class BIT_STRING_BITNAME_st(Structure):
    pass
BIT_STRING_BITNAME = BIT_STRING_BITNAME_st
asn1_output_data_fn = CFUNCTYPE(c_int, POINTER(BIO), POINTER(BIO), POINTER(ASN1_VALUE), c_int, POINTER(ASN1_ITEM))
class obj_name_st(Structure):
    pass
OBJ_NAME = obj_name_st
evp_sign_method = CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), c_uint, POINTER(c_ubyte), POINTER(c_uint), c_void_p)
evp_verify_method = CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), c_uint, POINTER(c_ubyte), c_uint, c_void_p)
class evp_cipher_info_st(Structure):
    pass
EVP_CIPHER_INFO = evp_cipher_info_st
class evp_Encode_Ctx_st(Structure):
    pass
EVP_ENCODE_CTX = evp_Encode_Ctx_st
EVP_PBE_KEYGEN = CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX), STRING, c_int, POINTER(ASN1_TYPE), POINTER(EVP_CIPHER), POINTER(EVP_MD), c_int)
class aes_key_st(Structure):
    pass
AES_KEY = aes_key_st
class rc4_key_st(Structure):
    pass
RC4_KEY = rc4_key_st
class bf_key_st(Structure):
    pass
BF_KEY = bf_key_st
DES_cblock = c_ubyte * 8
const_DES_cblock = c_ubyte * 8
class DES_ks(Structure):
    pass
DES_key_schedule = DES_ks
_ossl_old_des_cblock = c_ubyte * 8
class _ossl_old_des_ks_struct(Structure):
    pass
class N23_ossl_old_des_ks_struct4DOT_31E(Union):
    pass
N23_ossl_old_des_ks_struct4DOT_31E._fields_ = [
    ('_', _ossl_old_des_cblock),
    ('pad', c_ulong * 2),
]
_ossl_old_des_ks_struct._fields_ = [
    ('ks', N23_ossl_old_des_ks_struct4DOT_31E),
]
_ossl_old_des_key_schedule = _ossl_old_des_ks_struct * 16
class ui_string_st(Structure):
    pass
UI_STRING = ui_string_st
class cast_key_st(Structure):
    pass
CAST_KEY = cast_key_st
class hmac_ctx_st(Structure):
    pass
HMAC_CTX = hmac_ctx_st
class DSA_SIG_st(Structure):
    pass
DSA_SIG = DSA_SIG_st
class ec_method_st(Structure):
    pass
EC_METHOD = ec_method_st
class ec_group_st(Structure):
    pass
EC_GROUP = ec_group_st
class ec_point_st(Structure):
    pass
EC_POINT = ec_point_st
class ecpk_parameters_st(Structure):
    pass
ECPKPARAMETERS = ecpk_parameters_st
class ec_key_st(Structure):
    pass
EC_KEY = ec_key_st
class ECDSA_SIG_st(Structure):
    pass
ECDSA_SIG = ECDSA_SIG_st
ptrdiff_t = c_int
class SHAstate_st(Structure):
    pass
SHA_CTX = SHAstate_st
class SHA256state_st(Structure):
    pass
SHA256_CTX = SHA256state_st
class SHA512state_st(Structure):
    pass
SHA512_CTX = SHA512state_st
class X509_objects_st(Structure):
    pass
X509_OBJECTS = X509_objects_st
X509_ALGORS = STACK
class X509_val_st(Structure):
    pass
X509_VAL = X509_val_st
class X509_pubkey_st(Structure):
    pass
X509_PUBKEY = X509_pubkey_st
class X509_sig_st(Structure):
    pass
X509_SIG = X509_sig_st
class X509_name_entry_st(Structure):
    pass
X509_NAME_ENTRY = X509_name_entry_st
class X509_extension_st(Structure):
    pass
X509_EXTENSION = X509_extension_st
X509_EXTENSIONS = STACK
class x509_attributes_st(Structure):
    pass
X509_ATTRIBUTE = x509_attributes_st
class X509_req_info_st(Structure):
    pass
X509_REQ_INFO = X509_req_info_st
class X509_req_st(Structure):
    pass
X509_REQ = X509_req_st
class x509_cinf_st(Structure):
    pass
X509_CINF = x509_cinf_st
class x509_cert_aux_st(Structure):
    pass
X509_CERT_AUX = x509_cert_aux_st
class x509_trust_st(Structure):
    pass
X509_TRUST = x509_trust_st
class x509_cert_pair_st(Structure):
    pass
X509_CERT_PAIR = x509_cert_pair_st
class X509_revoked_st(Structure):
    pass
X509_REVOKED = X509_revoked_st
class X509_crl_info_st(Structure):
    pass
X509_CRL_INFO = X509_crl_info_st
class private_key_st(Structure):
    pass
X509_PKEY = private_key_st
class X509_info_st(Structure):
    pass
X509_INFO = X509_info_st
class Netscape_spkac_st(Structure):
    pass
NETSCAPE_SPKAC = Netscape_spkac_st
class Netscape_spki_st(Structure):
    pass
NETSCAPE_SPKI = Netscape_spki_st
class Netscape_certificate_sequence(Structure):
    pass
NETSCAPE_CERT_SEQUENCE = Netscape_certificate_sequence
class PBEPARAM_st(Structure):
    pass
PBEPARAM = PBEPARAM_st
class PBE2PARAM_st(Structure):
    pass
PBE2PARAM = PBE2PARAM_st
class PBKDF2PARAM_st(Structure):
    pass
PBKDF2PARAM = PBKDF2PARAM_st
class pkcs8_priv_key_info_st(Structure):
    pass
PKCS8_PRIV_KEY_INFO = pkcs8_priv_key_info_st
class lhash_node_st(Structure):
    pass
LHASH_NODE = lhash_node_st
LHASH_COMP_FN_TYPE = CFUNCTYPE(c_int, c_void_p, c_void_p)
LHASH_HASH_FN_TYPE = CFUNCTYPE(c_ulong, c_void_p)
LHASH_DOALL_FN_TYPE = CFUNCTYPE(None, c_void_p)
LHASH_DOALL_ARG_FN_TYPE = CFUNCTYPE(None, c_void_p, c_void_p)
class lhash_st(Structure):
    pass
LHASH = lhash_st
class x509_hash_dir_st(Structure):
    pass
X509_HASH_DIR_CTX = x509_hash_dir_st
class x509_file_st(Structure):
    pass
X509_CERT_FILE_CTX = x509_file_st
class x509_object_st(Structure):
    pass
X509_OBJECT = x509_object_st
class x509_lookup_st(Structure):
    pass
X509_LOOKUP = x509_lookup_st
class x509_lookup_method_st(Structure):
    pass
X509_LOOKUP_METHOD = x509_lookup_method_st
class X509_VERIFY_PARAM_st(Structure):
    pass
X509_VERIFY_PARAM = X509_VERIFY_PARAM_st
class pkcs7_issuer_and_serial_st(Structure):
    pass
PKCS7_ISSUER_AND_SERIAL = pkcs7_issuer_and_serial_st
class pkcs7_signer_info_st(Structure):
    pass
PKCS7_SIGNER_INFO = pkcs7_signer_info_st
class pkcs7_recip_info_st(Structure):
    pass
PKCS7_RECIP_INFO = pkcs7_recip_info_st
class pkcs7_signed_st(Structure):
    pass
PKCS7_SIGNED = pkcs7_signed_st
class pkcs7_enc_content_st(Structure):
    pass
PKCS7_ENC_CONTENT = pkcs7_enc_content_st
class pkcs7_enveloped_st(Structure):
    pass
PKCS7_ENVELOPE = pkcs7_enveloped_st
class pkcs7_signedandenveloped_st(Structure):
    pass
PKCS7_SIGN_ENVELOPE = pkcs7_signedandenveloped_st
class pkcs7_digest_st(Structure):
    pass
PKCS7_DIGEST = pkcs7_digest_st
class pkcs7_encrypted_st(Structure):
    pass
PKCS7_ENCRYPT = pkcs7_encrypted_st
class pkcs7_st(Structure):
    pass
PKCS7 = pkcs7_st

# values for enumeration 'STORE_object_types'
STORE_OBJECT_TYPE_X509_CERTIFICATE = 1
STORE_OBJECT_TYPE_X509_CRL = 2
STORE_OBJECT_TYPE_PRIVATE_KEY = 3
STORE_OBJECT_TYPE_PUBLIC_KEY = 4
STORE_OBJECT_TYPE_NUMBER = 5
STORE_OBJECT_TYPE_ARBITRARY = 6
STORE_OBJECT_TYPE_NUM = 6
STORE_object_types = c_int # enum
STORE_OBJECT_TYPES = STORE_object_types

# values for enumeration 'STORE_params'
STORE_PARAM_EVP_TYPE = 1
STORE_PARAM_BITS = 2
STORE_PARAM_KEY_PARAMETERS = 3
STORE_PARAM_KEY_NO_PARAMETERS = 4
STORE_PARAM_AUTH_PASSPHRASE = 5
STORE_PARAM_AUTH_KRB5_TICKET = 6
STORE_PARAM_TYPE_NUM = 6
STORE_params = c_int # enum
STORE_PARAM_TYPES = STORE_params

# values for enumeration 'STORE_attribs'
STORE_ATTR_END = 0
STORE_ATTR_FRIENDLYNAME = 1
STORE_ATTR_KEYID = 2
STORE_ATTR_ISSUERKEYID = 3
STORE_ATTR_SUBJECTKEYID = 4
STORE_ATTR_ISSUERSERIALHASH = 5
STORE_ATTR_ISSUER = 6
STORE_ATTR_SERIAL = 7
STORE_ATTR_SUBJECT = 8
STORE_ATTR_CERTHASH = 9
STORE_ATTR_EMAIL = 10
STORE_ATTR_FILENAME = 11
STORE_ATTR_TYPE_NUM = 11
STORE_ATTR_OR = 255
STORE_attribs = c_int # enum
STORE_ATTR_TYPES = STORE_attribs

# values for enumeration 'STORE_certificate_status'
STORE_X509_VALID = 0
STORE_X509_EXPIRED = 1
STORE_X509_SUSPENDED = 2
STORE_X509_REVOKED = 3
STORE_certificate_status = c_int # enum
STORE_CERTIFICATE_STATUS = STORE_certificate_status
class STORE_OBJECT_st(Structure):
    pass
STORE_OBJECT = STORE_OBJECT_st
STORE_INITIALISE_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE))
STORE_CLEANUP_FUNC_PTR = CFUNCTYPE(None, POINTER(STORE))
STORE_GENERATE_OBJECT_FUNC_PTR = CFUNCTYPE(POINTER(STORE_OBJECT), POINTER(STORE), STORE_OBJECT_TYPES, POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_GET_OBJECT_FUNC_PTR = CFUNCTYPE(POINTER(STORE_OBJECT), POINTER(STORE), STORE_OBJECT_TYPES, POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_START_OBJECT_FUNC_PTR = CFUNCTYPE(c_void_p, POINTER(STORE), STORE_OBJECT_TYPES, POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_NEXT_OBJECT_FUNC_PTR = CFUNCTYPE(POINTER(STORE_OBJECT), POINTER(STORE), c_void_p)
STORE_END_OBJECT_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE), c_void_p)
STORE_HANDLE_OBJECT_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE), STORE_OBJECT_TYPES, POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_STORE_OBJECT_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE), STORE_OBJECT_TYPES, POINTER(STORE_OBJECT), POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_MODIFY_OBJECT_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE), STORE_OBJECT_TYPES, POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_GENERIC_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE), POINTER(OPENSSL_ITEM), POINTER(OPENSSL_ITEM))
STORE_CTRL_FUNC_PTR = CFUNCTYPE(c_int, POINTER(STORE), c_int, c_long, c_void_p, CFUNCTYPE(None))
class STORE_attr_info_st(Structure):
    pass
STORE_ATTR_INFO = STORE_attr_info_st
error_t = c_int
class err_state_st(Structure):
    pass
ERR_STATE = err_state_st
class ERR_string_data_st(Structure):
    pass
ERR_STRING_DATA = ERR_string_data_st
class ENGINE_CMD_DEFN_st(Structure):
    pass
ENGINE_CMD_DEFN = ENGINE_CMD_DEFN_st
ENGINE_GEN_FUNC_PTR = CFUNCTYPE(c_int)
ENGINE_GEN_INT_FUNC_PTR = CFUNCTYPE(c_int, POINTER(ENGINE))
ENGINE_CTRL_FUNC_PTR = CFUNCTYPE(c_int, POINTER(ENGINE), c_int, c_long, c_void_p, CFUNCTYPE(None))
ENGINE_LOAD_KEY_PTR = CFUNCTYPE(POINTER(EVP_PKEY), POINTER(ENGINE), STRING, POINTER(UI_METHOD), c_void_p)
ENGINE_SSL_CLIENT_CERT_PTR = CFUNCTYPE(c_int, POINTER(ENGINE), POINTER(SSL), POINTER(STACK), POINTER(POINTER(X509)), POINTER(POINTER(EVP_PKEY)), POINTER(POINTER(STACK)), POINTER(UI_METHOD), c_void_p)
ENGINE_CIPHERS_PTR = CFUNCTYPE(c_int, POINTER(ENGINE), POINTER(POINTER(EVP_CIPHER)), POINTER(POINTER(c_int)), c_int)
ENGINE_DIGESTS_PTR = CFUNCTYPE(c_int, POINTER(ENGINE), POINTER(POINTER(EVP_MD)), POINTER(POINTER(c_int)), c_int)
dyn_MEM_malloc_cb = CFUNCTYPE(c_void_p, size_t)
dyn_MEM_realloc_cb = CFUNCTYPE(c_void_p, c_void_p, size_t)
dyn_MEM_free_cb = CFUNCTYPE(None, c_void_p)
class st_dynamic_MEM_fns(Structure):
    pass
dynamic_MEM_fns = st_dynamic_MEM_fns
dyn_lock_locking_cb = CFUNCTYPE(None, c_int, c_int, STRING, c_int)
dyn_lock_add_lock_cb = CFUNCTYPE(c_int, POINTER(c_int), c_int, c_int, STRING, c_int)
class CRYPTO_dynlock_value(Structure):
    pass
dyn_dynlock_create_cb = CFUNCTYPE(POINTER(CRYPTO_dynlock_value), STRING, c_int)
dyn_dynlock_lock_cb = CFUNCTYPE(None, c_int, POINTER(CRYPTO_dynlock_value), STRING, c_int)
dyn_dynlock_destroy_cb = CFUNCTYPE(None, POINTER(CRYPTO_dynlock_value), STRING, c_int)
class st_dynamic_LOCK_fns(Structure):
    pass
dynamic_LOCK_fns = st_dynamic_LOCK_fns
class st_dynamic_fns(Structure):
    pass
dynamic_fns = st_dynamic_fns
dynamic_v_check_fn = CFUNCTYPE(c_ulong, c_ulong)
dynamic_bind_engine = CFUNCTYPE(c_int, POINTER(ENGINE), STRING, POINTER(dynamic_fns))
bignum_ctx._fields_ = [
]
bn_blinding_st._fields_ = [
]
ecdh_method._fields_ = [
]
ecdsa_method._fields_ = [
]
ssl_st._fields_ = [
]
ssl_ctx_st._fields_ = [
]
v3_ext_ctx._fields_ = [
]
conf_st._fields_ = [
]
store_st._fields_ = [
]
store_method_st._fields_ = [
]
ui_st._fields_ = [
]
ui_method_st._fields_ = [
]
st_ERR_FNS._fields_ = [
]
engine_st._fields_ = [
]
X509_POLICY_NODE_st._fields_ = [
]
X509_POLICY_LEVEL_st._fields_ = [
]
X509_POLICY_TREE_st._fields_ = [
]
X509_POLICY_CACHE_st._fields_ = [
]
ocsp_req_ctx_st._fields_ = [
]
ocsp_response_st._fields_ = [
]
ocsp_responder_id_st._fields_ = [
]
__fsid_t._fields_ = [
    ('__val', c_int * 2),
]
class __mbstate_t(Structure):
    pass
class N11__mbstate_t3DOT_2E(Union):
    pass
N11__mbstate_t3DOT_2E._fields_ = [
    ('__wch', c_uint),
    ('__wchb', c_char * 4),
]
__mbstate_t._fields_ = [
    ('__count', c_int),
    ('__value', N11__mbstate_t3DOT_2E),
]
_G_fpos_t._fields_ = [
    ('__pos', __off_t),
    ('__state', __mbstate_t),
]
_G_fpos64_t._pack_ = 4
_G_fpos64_t._fields_ = [
    ('__pos', __off64_t),
    ('__state', __mbstate_t),
]
class _IO_marker(Structure):
    pass
_IO_FILE._pack_ = 4
_IO_FILE._fields_ = [
    ('_flags', c_int),
    ('_IO_read_ptr', STRING),
    ('_IO_read_end', STRING),
    ('_IO_read_base', STRING),
    ('_IO_write_base', STRING),
    ('_IO_write_ptr', STRING),
    ('_IO_write_end', STRING),
    ('_IO_buf_base', STRING),
    ('_IO_buf_end', STRING),
    ('_IO_save_base', STRING),
    ('_IO_backup_base', STRING),
    ('_IO_save_end', STRING),
    ('_markers', POINTER(_IO_marker)),
    ('_chain', POINTER(_IO_FILE)),
    ('_fileno', c_int),
    ('_flags2', c_int),
    ('_old_offset', __off_t),
    ('_cur_column', c_ushort),
    ('_vtable_offset', c_byte),
    ('_shortbuf', c_char * 1),
    ('_lock', POINTER(_IO_lock_t)),
    ('_offset', __off64_t),
    ('__pad1', c_void_p),
    ('__pad2', c_void_p),
    ('__pad3', c_void_p),
    ('__pad4', c_void_p),
    ('__pad5', size_t),
    ('_mode', c_int),
    ('_unused2', c_char * 40),
]
_IO_cookie_io_functions_t._fields_ = [
    ('read', POINTER(__io_read_fn)),
    ('write', POINTER(__io_write_fn)),
    ('seek', POINTER(__io_seek_fn)),
    ('close', POINTER(__io_close_fn)),
]
class __locale_data(Structure):
    pass
__locale_struct._fields_ = [
    ('__locales', POINTER(__locale_data) * 13),
    ('__ctype_b', POINTER(c_ushort)),
    ('__ctype_tolower', POINTER(c_int)),
    ('__ctype_toupper', POINTER(c_int)),
    ('__names', STRING * 13),
]
__sigset_t._fields_ = [
    ('__val', c_ulong * 32),
]
__pthread_internal_slist._fields_ = [
    ('__next', POINTER(__pthread_internal_slist)),
]
stack_st._fields_ = [
    ('num', c_int),
    ('data', POINTER(STRING)),
    ('sorted', c_int),
    ('num_alloc', c_int),
    ('comp', CFUNCTYPE(c_int, POINTER(STRING), POINTER(STRING))),
]
openssl_item_st._fields_ = [
    ('code', c_int),
    ('value', c_void_p),
    ('value_size', size_t),
    ('value_length', POINTER(size_t)),
]
CRYPTO_dynlock_value._fields_ = [
]
crypto_ex_data_st._fields_ = [
    ('sk', POINTER(STACK)),
    ('dummy', c_int),
]
crypto_ex_data_func_st._fields_ = [
    ('argl', c_long),
    ('argp', c_void_p),
    ('new_func', POINTER(CRYPTO_EX_new)),
    ('free_func', POINTER(CRYPTO_EX_free)),
    ('dup_func', POINTER(CRYPTO_EX_dup)),
]
st_CRYPTO_EX_DATA_IMPL._fields_ = [
]
bio_method_st._fields_ = [
    ('type', c_int),
    ('name', STRING),
    ('bwrite', CFUNCTYPE(c_int, POINTER(BIO), STRING, c_int)),
    ('bread', CFUNCTYPE(c_int, POINTER(BIO), STRING, c_int)),
    ('bputs', CFUNCTYPE(c_int, POINTER(BIO), STRING)),
    ('bgets', CFUNCTYPE(c_int, POINTER(BIO), STRING, c_int)),
    ('ctrl', CFUNCTYPE(c_long, POINTER(BIO), c_int, c_long, c_void_p)),
    ('create', CFUNCTYPE(c_int, POINTER(BIO))),
    ('destroy', CFUNCTYPE(c_int, POINTER(BIO))),
    ('callback_ctrl', CFUNCTYPE(c_long, POINTER(BIO), c_int, POINTER(bio_info_cb))),
]
bio_st._fields_ = [
    ('method', POINTER(BIO_METHOD)),
    ('callback', CFUNCTYPE(c_long, POINTER(bio_st), c_int, STRING, c_int, c_long, c_long)),
    ('cb_arg', STRING),
    ('init', c_int),
    ('shutdown', c_int),
    ('flags', c_int),
    ('retry_reason', c_int),
    ('num', c_int),
    ('ptr', c_void_p),
    ('next_bio', POINTER(bio_st)),
    ('prev_bio', POINTER(bio_st)),
    ('references', c_int),
    ('num_read', c_ulong),
    ('num_write', c_ulong),
    ('ex_data', CRYPTO_EX_DATA),
]
bio_f_buffer_ctx_struct._fields_ = [
    ('ibuf_size', c_int),
    ('obuf_size', c_int),
    ('ibuf', STRING),
    ('ibuf_len', c_int),
    ('ibuf_off', c_int),
    ('obuf', STRING),
    ('obuf_len', c_int),
    ('obuf_off', c_int),
]
bignum_st._fields_ = [
    ('d', POINTER(c_ulong)),
    ('top', c_int),
    ('dmax', c_int),
    ('neg', c_int),
    ('flags', c_int),
]
bn_mont_ctx_st._fields_ = [
    ('ri', c_int),
    ('RR', BIGNUM),
    ('N', BIGNUM),
    ('Ni', BIGNUM),
    ('n0', c_ulong),
    ('flags', c_int),
]
bn_recp_ctx_st._fields_ = [
    ('N', BIGNUM),
    ('Nr', BIGNUM),
    ('num_bits', c_int),
    ('shift', c_int),
    ('flags', c_int),
]
class N11bn_gencb_st4DOT_26E(Union):
    pass
N11bn_gencb_st4DOT_26E._fields_ = [
    ('cb_1', CFUNCTYPE(None, c_int, c_int, c_void_p)),
    ('cb_2', CFUNCTYPE(c_int, c_int, c_int, POINTER(BN_GENCB))),
]
bn_gencb_st._fields_ = [
    ('ver', c_uint),
    ('arg', c_void_p),
    ('cb', N11bn_gencb_st4DOT_26E),
]
asn1_ctx_st._fields_ = [
    ('p', POINTER(c_ubyte)),
    ('eos', c_int),
    ('error', c_int),
    ('inf', c_int),
    ('tag', c_int),
    ('xclass', c_int),
    ('slen', c_long),
    ('max', POINTER(c_ubyte)),
    ('q', POINTER(c_ubyte)),
    ('pp', POINTER(POINTER(c_ubyte))),
    ('line', c_int),
]
asn1_const_ctx_st._fields_ = [
    ('p', POINTER(c_ubyte)),
    ('eos', c_int),
    ('error', c_int),
    ('inf', c_int),
    ('tag', c_int),
    ('xclass', c_int),
    ('slen', c_long),
    ('max', POINTER(c_ubyte)),
    ('q', POINTER(c_ubyte)),
    ('pp', POINTER(POINTER(c_ubyte))),
    ('line', c_int),
]
asn1_object_st._fields_ = [
    ('sn', STRING),
    ('ln', STRING),
    ('nid', c_int),
    ('length', c_int),
    ('data', POINTER(c_ubyte)),
    ('flags', c_int),
]
asn1_string_st._fields_ = [
    ('length', c_int),
    ('type', c_int),
    ('data', POINTER(c_ubyte)),
    ('flags', c_long),
]
ASN1_ENCODING_st._fields_ = [
    ('enc', POINTER(c_ubyte)),
    ('len', c_long),
    ('modified', c_int),
]
asn1_string_table_st._fields_ = [
    ('nid', c_int),
    ('minsize', c_long),
    ('maxsize', c_long),
    ('mask', c_ulong),
    ('flags', c_ulong),
]
ASN1_TEMPLATE_st._fields_ = [
]
ASN1_ITEM_st._fields_ = [
]
ASN1_TLC_st._fields_ = [
]
ASN1_VALUE_st._fields_ = [
]
class N12asn1_type_st4DOT_27E(Union):
    pass
N12asn1_type_st4DOT_27E._fields_ = [
    ('ptr', STRING),
    ('boolean', ASN1_BOOLEAN),
    ('asn1_string', POINTER(ASN1_STRING)),
    ('object', POINTER(ASN1_OBJECT)),
    ('integer', POINTER(ASN1_INTEGER)),
    ('enumerated', POINTER(ASN1_ENUMERATED)),
    ('bit_string', POINTER(ASN1_BIT_STRING)),
    ('octet_string', POINTER(ASN1_OCTET_STRING)),
    ('printablestring', POINTER(ASN1_PRINTABLESTRING)),
    ('t61string', POINTER(ASN1_T61STRING)),
    ('ia5string', POINTER(ASN1_IA5STRING)),
    ('generalstring', POINTER(ASN1_GENERALSTRING)),
    ('bmpstring', POINTER(ASN1_BMPSTRING)),
    ('universalstring', POINTER(ASN1_UNIVERSALSTRING)),
    ('utctime', POINTER(ASN1_UTCTIME)),
    ('generalizedtime', POINTER(ASN1_GENERALIZEDTIME)),
    ('visiblestring', POINTER(ASN1_VISIBLESTRING)),
    ('utf8string', POINTER(ASN1_UTF8STRING)),
    ('set', POINTER(ASN1_STRING)),
    ('sequence', POINTER(ASN1_STRING)),
    ('asn1_value', POINTER(ASN1_VALUE)),
]
asn1_type_st._fields_ = [
    ('type', c_int),
    ('value', N12asn1_type_st4DOT_27E),
]
asn1_method_st._fields_ = [
    ('i2d', POINTER(i2d_of_void)),
    ('d2i', POINTER(d2i_of_void)),
    ('create', CFUNCTYPE(c_void_p)),
    ('destroy', CFUNCTYPE(None, c_void_p)),
]
asn1_header_st._fields_ = [
    ('header', POINTER(ASN1_OCTET_STRING)),
    ('data', c_void_p),
    ('meth', POINTER(ASN1_METHOD)),
]
BIT_STRING_BITNAME_st._fields_ = [
    ('bitnum', c_int),
    ('lname', STRING),
    ('sname', STRING),
]
obj_name_st._fields_ = [
    ('type', c_int),
    ('alias', c_int),
    ('name', STRING),
    ('data', STRING),
]
class N11evp_pkey_st4DOT_28E(Union):
    pass
N11evp_pkey_st4DOT_28E._fields_ = [
    ('ptr', STRING),
    ('rsa', POINTER(rsa_st)),
    ('dsa', POINTER(dsa_st)),
    ('dh', POINTER(dh_st)),
    ('ec', POINTER(ec_key_st)),
]
evp_pkey_st._fields_ = [
    ('type', c_int),
    ('save_type', c_int),
    ('references', c_int),
    ('pkey', N11evp_pkey_st4DOT_28E),
    ('save_parameters', c_int),
    ('attributes', POINTER(STACK)),
]
ec_key_st._fields_ = [
]
env_md_st._fields_ = [
    ('type', c_int),
    ('pkey_type', c_int),
    ('md_size', c_int),
    ('flags', c_ulong),
    ('init', CFUNCTYPE(c_int, POINTER(EVP_MD_CTX))),
    ('update', CFUNCTYPE(c_int, POINTER(EVP_MD_CTX), c_void_p, size_t)),
    ('final', CFUNCTYPE(c_int, POINTER(EVP_MD_CTX), POINTER(c_ubyte))),
    ('copy', CFUNCTYPE(c_int, POINTER(EVP_MD_CTX), POINTER(EVP_MD_CTX))),
    ('cleanup', CFUNCTYPE(c_int, POINTER(EVP_MD_CTX))),
    ('sign', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), c_uint, POINTER(c_ubyte), POINTER(c_uint), c_void_p)),
    ('verify', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), c_uint, POINTER(c_ubyte), c_uint, c_void_p)),
    ('required_pkey_type', c_int * 5),
    ('block_size', c_int),
    ('ctx_size', c_int),
]
env_md_ctx_st._fields_ = [
    ('digest', POINTER(EVP_MD)),
    ('engine', POINTER(ENGINE)),
    ('flags', c_ulong),
    ('md_data', c_void_p),
]
evp_cipher_st._fields_ = [
    ('nid', c_int),
    ('block_size', c_int),
    ('key_len', c_int),
    ('iv_len', c_int),
    ('flags', c_ulong),
    ('init', CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX), POINTER(c_ubyte), POINTER(c_ubyte), c_int)),
    ('do_cipher', CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX), POINTER(c_ubyte), POINTER(c_ubyte), c_uint)),
    ('cleanup', CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX))),
    ('ctx_size', c_int),
    ('set_asn1_parameters', CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX), POINTER(ASN1_TYPE))),
    ('get_asn1_parameters', CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX), POINTER(ASN1_TYPE))),
    ('ctrl', CFUNCTYPE(c_int, POINTER(EVP_CIPHER_CTX), c_int, c_int, c_void_p)),
    ('app_data', c_void_p),
]
evp_cipher_info_st._fields_ = [
    ('cipher', POINTER(EVP_CIPHER)),
    ('iv', c_ubyte * 16),
]
evp_cipher_ctx_st._fields_ = [
    ('cipher', POINTER(EVP_CIPHER)),
    ('engine', POINTER(ENGINE)),
    ('encrypt', c_int),
    ('buf_len', c_int),
    ('oiv', c_ubyte * 16),
    ('iv', c_ubyte * 16),
    ('buf', c_ubyte * 32),
    ('num', c_int),
    ('app_data', c_void_p),
    ('key_len', c_int),
    ('flags', c_ulong),
    ('cipher_data', c_void_p),
    ('final_used', c_int),
    ('block_mask', c_int),
    ('final', c_ubyte * 32),
]
evp_Encode_Ctx_st._fields_ = [
    ('num', c_int),
    ('length', c_int),
    ('enc_data', c_ubyte * 80),
    ('line_num', c_int),
    ('expect_nl', c_int),
]
aes_key_st._fields_ = [
    ('rd_key', c_uint * 60),
    ('rounds', c_int),
]
rc4_key_st._fields_ = [
    ('x', c_uint),
    ('y', c_uint),
    ('data', c_uint * 256),
]
bf_key_st._fields_ = [
    ('P', c_uint * 18),
    ('S', c_uint * 1024),
]
class N6DES_ks4DOT_30E(Union):
    pass
N6DES_ks4DOT_30E._fields_ = [
    ('cblock', DES_cblock),
    ('deslong', c_ulong * 2),
]
DES_ks._fields_ = [
    ('ks', N6DES_ks4DOT_30E * 16),
]
ui_string_st._fields_ = [
]
cast_key_st._fields_ = [
    ('data', c_ulong * 32),
    ('short_key', c_int),
]
hmac_ctx_st._fields_ = [
    ('md', POINTER(EVP_MD)),
    ('md_ctx', EVP_MD_CTX),
    ('i_ctx', EVP_MD_CTX),
    ('o_ctx', EVP_MD_CTX),
    ('key_length', c_uint),
    ('key', c_ubyte * 128),
]
dh_method._fields_ = [
    ('name', STRING),
    ('generate_key', CFUNCTYPE(c_int, POINTER(DH))),
    ('compute_key', CFUNCTYPE(c_int, POINTER(c_ubyte), POINTER(BIGNUM), POINTER(DH))),
    ('bn_mod_exp', CFUNCTYPE(c_int, POINTER(DH), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BN_CTX), POINTER(BN_MONT_CTX))),
    ('init', CFUNCTYPE(c_int, POINTER(DH))),
    ('finish', CFUNCTYPE(c_int, POINTER(DH))),
    ('flags', c_int),
    ('app_data', STRING),
    ('generate_params', CFUNCTYPE(c_int, POINTER(DH), c_int, c_int, POINTER(BN_GENCB))),
]
dh_st._fields_ = [
    ('pad', c_int),
    ('version', c_int),
    ('p', POINTER(BIGNUM)),
    ('g', POINTER(BIGNUM)),
    ('length', c_long),
    ('pub_key', POINTER(BIGNUM)),
    ('priv_key', POINTER(BIGNUM)),
    ('flags', c_int),
    ('method_mont_p', POINTER(BN_MONT_CTX)),
    ('q', POINTER(BIGNUM)),
    ('j', POINTER(BIGNUM)),
    ('seed', POINTER(c_ubyte)),
    ('seedlen', c_int),
    ('counter', POINTER(BIGNUM)),
    ('references', c_int),
    ('ex_data', CRYPTO_EX_DATA),
    ('meth', POINTER(DH_METHOD)),
    ('engine', POINTER(ENGINE)),
]
DSA_SIG_st._fields_ = [
    ('r', POINTER(BIGNUM)),
    ('s', POINTER(BIGNUM)),
]
dsa_method._fields_ = [
    ('name', STRING),
    ('dsa_do_sign', CFUNCTYPE(POINTER(DSA_SIG), POINTER(c_ubyte), c_int, POINTER(DSA))),
    ('dsa_sign_setup', CFUNCTYPE(c_int, POINTER(DSA), POINTER(BN_CTX), POINTER(POINTER(BIGNUM)), POINTER(POINTER(BIGNUM)))),
    ('dsa_do_verify', CFUNCTYPE(c_int, POINTER(c_ubyte), c_int, POINTER(DSA_SIG), POINTER(DSA))),
    ('dsa_mod_exp', CFUNCTYPE(c_int, POINTER(DSA), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BN_CTX), POINTER(BN_MONT_CTX))),
    ('bn_mod_exp', CFUNCTYPE(c_int, POINTER(DSA), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BN_CTX), POINTER(BN_MONT_CTX))),
    ('init', CFUNCTYPE(c_int, POINTER(DSA))),
    ('finish', CFUNCTYPE(c_int, POINTER(DSA))),
    ('flags', c_int),
    ('app_data', STRING),
    ('dsa_paramgen', CFUNCTYPE(c_int, POINTER(DSA), c_int, POINTER(c_ubyte), c_int, POINTER(c_int), POINTER(c_ulong), POINTER(BN_GENCB))),
    ('dsa_keygen', CFUNCTYPE(c_int, POINTER(DSA))),
]
dsa_st._fields_ = [
    ('pad', c_int),
    ('version', c_long),
    ('write_params', c_int),
    ('p', POINTER(BIGNUM)),
    ('q', POINTER(BIGNUM)),
    ('g', POINTER(BIGNUM)),
    ('pub_key', POINTER(BIGNUM)),
    ('priv_key', POINTER(BIGNUM)),
    ('kinv', POINTER(BIGNUM)),
    ('r', POINTER(BIGNUM)),
    ('flags', c_int),
    ('method_mont_p', POINTER(BN_MONT_CTX)),
    ('references', c_int),
    ('ex_data', CRYPTO_EX_DATA),
    ('meth', POINTER(DSA_METHOD)),
    ('engine', POINTER(ENGINE)),
]
rsa_meth_st._fields_ = [
    ('name', STRING),
    ('rsa_pub_enc', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), POINTER(c_ubyte), POINTER(RSA), c_int)),
    ('rsa_pub_dec', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), POINTER(c_ubyte), POINTER(RSA), c_int)),
    ('rsa_priv_enc', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), POINTER(c_ubyte), POINTER(RSA), c_int)),
    ('rsa_priv_dec', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), POINTER(c_ubyte), POINTER(RSA), c_int)),
    ('rsa_mod_exp', CFUNCTYPE(c_int, POINTER(BIGNUM), POINTER(BIGNUM), POINTER(RSA), POINTER(BN_CTX))),
    ('bn_mod_exp', CFUNCTYPE(c_int, POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BIGNUM), POINTER(BN_CTX), POINTER(BN_MONT_CTX))),
    ('init', CFUNCTYPE(c_int, POINTER(RSA))),
    ('finish', CFUNCTYPE(c_int, POINTER(RSA))),
    ('flags', c_int),
    ('app_data', STRING),
    ('rsa_sign', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), c_uint, POINTER(c_ubyte), POINTER(c_uint), POINTER(RSA))),
    ('rsa_verify', CFUNCTYPE(c_int, c_int, POINTER(c_ubyte), c_uint, POINTER(c_ubyte), c_uint, POINTER(RSA))),
    ('rsa_keygen', CFUNCTYPE(c_int, POINTER(RSA), c_int, POINTER(BIGNUM), POINTER(BN_GENCB))),
]
rsa_st._fields_ = [
    ('pad', c_int),
    ('version', c_long),
    ('meth', POINTER(RSA_METHOD)),
    ('engine', POINTER(ENGINE)),
    ('n', POINTER(BIGNUM)),
    ('e', POINTER(BIGNUM)),
    ('d', POINTER(BIGNUM)),
    ('p', POINTER(BIGNUM)),
    ('q', POINTER(BIGNUM)),
    ('dmp1', POINTER(BIGNUM)),
    ('dmq1', POINTER(BIGNUM)),
    ('iqmp', POINTER(BIGNUM)),
    ('ex_data', CRYPTO_EX_DATA),
    ('references', c_int),
    ('flags', c_int),
    ('_method_mod_n', POINTER(BN_MONT_CTX)),
    ('_method_mod_p', POINTER(BN_MONT_CTX)),
    ('_method_mod_q', POINTER(BN_MONT_CTX)),
    ('bignum_data', STRING),
    ('blinding', POINTER(BN_BLINDING)),
    ('mt_blinding', POINTER(BN_BLINDING)),
]
ec_method_st._fields_ = [
]
ec_group_st._fields_ = [
]
ec_point_st._fields_ = [
]
ecpk_parameters_st._fields_ = [
]
ECDSA_SIG_st._fields_ = [
    ('r', POINTER(BIGNUM)),
    ('s', POINTER(BIGNUM)),
]
rand_meth_st._fields_ = [
    ('seed', CFUNCTYPE(None, c_void_p, c_int)),
    ('bytes', CFUNCTYPE(c_int, POINTER(c_ubyte), c_int)),
    ('cleanup', CFUNCTYPE(None)),
    ('add', CFUNCTYPE(None, c_void_p, c_int, c_double)),
    ('pseudorand', CFUNCTYPE(c_int, POINTER(c_ubyte), c_int)),
    ('status', CFUNCTYPE(c_int)),
]
buf_mem_st._fields_ = [
    ('length', c_int),
    ('data', STRING),
    ('max', c_int),
]
SHAstate_st._fields_ = [
    ('h0', c_uint),
    ('h1', c_uint),
    ('h2', c_uint),
    ('h3', c_uint),
    ('h4', c_uint),
    ('Nl', c_uint),
    ('Nh', c_uint),
    ('data', c_uint * 16),
    ('num', c_uint),
]
SHA256state_st._fields_ = [
    ('h', c_uint * 8),
    ('Nl', c_uint),
    ('Nh', c_uint),
    ('data', c_uint * 16),
    ('num', c_uint),
    ('md_len', c_uint),
]
class N14SHA512state_st4DOT_34E(Union):
    pass
N14SHA512state_st4DOT_34E._pack_ = 4
N14SHA512state_st4DOT_34E._fields_ = [
    ('d', c_ulonglong * 16),
    ('p', c_ubyte * 128),
]
SHA512state_st._pack_ = 4
SHA512state_st._fields_ = [
    ('h', c_ulonglong * 8),
    ('Nl', c_ulonglong),
    ('Nh', c_ulonglong),
    ('u', N14SHA512state_st4DOT_34E),
    ('num', c_uint),
    ('md_len', c_uint),
]
X509_objects_st._fields_ = [
    ('nid', c_int),
    ('a2i', CFUNCTYPE(c_int)),
    ('i2a', CFUNCTYPE(c_int)),
]
X509_algor_st._fields_ = [
    ('algorithm', POINTER(ASN1_OBJECT)),
    ('parameter', POINTER(ASN1_TYPE)),
]
X509_val_st._fields_ = [
    ('notBefore', POINTER(ASN1_TIME)),
    ('notAfter', POINTER(ASN1_TIME)),
]
X509_pubkey_st._fields_ = [
    ('algor', POINTER(X509_ALGOR)),
    ('public_key', POINTER(ASN1_BIT_STRING)),
    ('pkey', POINTER(EVP_PKEY)),
]
X509_sig_st._fields_ = [
    ('algor', POINTER(X509_ALGOR)),
    ('digest', POINTER(ASN1_OCTET_STRING)),
]
X509_name_entry_st._fields_ = [
    ('object', POINTER(ASN1_OBJECT)),
    ('value', POINTER(ASN1_STRING)),
    ('set', c_int),
    ('size', c_int),
]
X509_name_st._fields_ = [
    ('entries', POINTER(STACK)),
    ('modified', c_int),
    ('bytes', POINTER(BUF_MEM)),
    ('hash', c_ulong),
]
X509_extension_st._fields_ = [
    ('object', POINTER(ASN1_OBJECT)),
    ('critical', ASN1_BOOLEAN),
    ('value', POINTER(ASN1_OCTET_STRING)),
]
class N18x509_attributes_st4DOT_35E(Union):
    pass
N18x509_attributes_st4DOT_35E._fields_ = [
    ('ptr', STRING),
    ('set', POINTER(STACK)),
    ('single', POINTER(ASN1_TYPE)),
]
x509_attributes_st._fields_ = [
    ('object', POINTER(ASN1_OBJECT)),
    ('single', c_int),
    ('value', N18x509_attributes_st4DOT_35E),
]
X509_req_info_st._fields_ = [
    ('enc', ASN1_ENCODING),
    ('version', POINTER(ASN1_INTEGER)),
    ('subject', POINTER(X509_NAME)),
    ('pubkey', POINTER(X509_PUBKEY)),
    ('attributes', POINTER(STACK)),
]
X509_req_st._fields_ = [
    ('req_info', POINTER(X509_REQ_INFO)),
    ('sig_alg', POINTER(X509_ALGOR)),
    ('signature', POINTER(ASN1_BIT_STRING)),
    ('references', c_int),
]
x509_cinf_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('serialNumber', POINTER(ASN1_INTEGER)),
    ('signature', POINTER(X509_ALGOR)),
    ('issuer', POINTER(X509_NAME)),
    ('validity', POINTER(X509_VAL)),
    ('subject', POINTER(X509_NAME)),
    ('key', POINTER(X509_PUBKEY)),
    ('issuerUID', POINTER(ASN1_BIT_STRING)),
    ('subjectUID', POINTER(ASN1_BIT_STRING)),
    ('extensions', POINTER(STACK)),
]
x509_cert_aux_st._fields_ = [
    ('trust', POINTER(STACK)),
    ('reject', POINTER(STACK)),
    ('alias', POINTER(ASN1_UTF8STRING)),
    ('keyid', POINTER(ASN1_OCTET_STRING)),
    ('other', POINTER(STACK)),
]
class AUTHORITY_KEYID_st(Structure):
    pass
x509_st._fields_ = [
    ('cert_info', POINTER(X509_CINF)),
    ('sig_alg', POINTER(X509_ALGOR)),
    ('signature', POINTER(ASN1_BIT_STRING)),
    ('valid', c_int),
    ('references', c_int),
    ('name', STRING),
    ('ex_data', CRYPTO_EX_DATA),
    ('ex_pathlen', c_long),
    ('ex_pcpathlen', c_long),
    ('ex_flags', c_ulong),
    ('ex_kusage', c_ulong),
    ('ex_xkusage', c_ulong),
    ('ex_nscert', c_ulong),
    ('skid', POINTER(ASN1_OCTET_STRING)),
    ('akid', POINTER(AUTHORITY_KEYID_st)),
    ('policy_cache', POINTER(X509_POLICY_CACHE)),
    ('sha1_hash', c_ubyte * 20),
    ('aux', POINTER(X509_CERT_AUX)),
]
x509_trust_st._fields_ = [
    ('trust', c_int),
    ('flags', c_int),
    ('check_trust', CFUNCTYPE(c_int, POINTER(x509_trust_st), POINTER(X509), c_int)),
    ('name', STRING),
    ('arg1', c_int),
    ('arg2', c_void_p),
]
x509_cert_pair_st._fields_ = [
    ('forward', POINTER(X509)),
    ('reverse', POINTER(X509)),
]
X509_revoked_st._fields_ = [
    ('serialNumber', POINTER(ASN1_INTEGER)),
    ('revocationDate', POINTER(ASN1_TIME)),
    ('extensions', POINTER(STACK)),
    ('sequence', c_int),
]
X509_crl_info_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('sig_alg', POINTER(X509_ALGOR)),
    ('issuer', POINTER(X509_NAME)),
    ('lastUpdate', POINTER(ASN1_TIME)),
    ('nextUpdate', POINTER(ASN1_TIME)),
    ('revoked', POINTER(STACK)),
    ('extensions', POINTER(STACK)),
    ('enc', ASN1_ENCODING),
]
X509_crl_st._fields_ = [
    ('crl', POINTER(X509_CRL_INFO)),
    ('sig_alg', POINTER(X509_ALGOR)),
    ('signature', POINTER(ASN1_BIT_STRING)),
    ('references', c_int),
]
private_key_st._fields_ = [
    ('version', c_int),
    ('enc_algor', POINTER(X509_ALGOR)),
    ('enc_pkey', POINTER(ASN1_OCTET_STRING)),
    ('dec_pkey', POINTER(EVP_PKEY)),
    ('key_length', c_int),
    ('key_data', STRING),
    ('key_free', c_int),
    ('cipher', EVP_CIPHER_INFO),
    ('references', c_int),
]
X509_info_st._fields_ = [
    ('x509', POINTER(X509)),
    ('crl', POINTER(X509_CRL)),
    ('x_pkey', POINTER(X509_PKEY)),
    ('enc_cipher', EVP_CIPHER_INFO),
    ('enc_len', c_int),
    ('enc_data', STRING),
    ('references', c_int),
]
Netscape_spkac_st._fields_ = [
    ('pubkey', POINTER(X509_PUBKEY)),
    ('challenge', POINTER(ASN1_IA5STRING)),
]
Netscape_spki_st._fields_ = [
    ('spkac', POINTER(NETSCAPE_SPKAC)),
    ('sig_algor', POINTER(X509_ALGOR)),
    ('signature', POINTER(ASN1_BIT_STRING)),
]
Netscape_certificate_sequence._fields_ = [
    ('type', POINTER(ASN1_OBJECT)),
    ('certs', POINTER(STACK)),
]
PBEPARAM_st._fields_ = [
    ('salt', POINTER(ASN1_OCTET_STRING)),
    ('iter', POINTER(ASN1_INTEGER)),
]
PBE2PARAM_st._fields_ = [
    ('keyfunc', POINTER(X509_ALGOR)),
    ('encryption', POINTER(X509_ALGOR)),
]
PBKDF2PARAM_st._fields_ = [
    ('salt', POINTER(ASN1_TYPE)),
    ('iter', POINTER(ASN1_INTEGER)),
    ('keylength', POINTER(ASN1_INTEGER)),
    ('prf', POINTER(X509_ALGOR)),
]
pkcs8_priv_key_info_st._fields_ = [
    ('broken', c_int),
    ('version', POINTER(ASN1_INTEGER)),
    ('pkeyalg', POINTER(X509_ALGOR)),
    ('pkey', POINTER(ASN1_TYPE)),
    ('attributes', POINTER(STACK)),
]
lhash_node_st._fields_ = [
    ('data', c_void_p),
    ('next', POINTER(lhash_node_st)),
    ('hash', c_ulong),
]
lhash_st._fields_ = [
    ('b', POINTER(POINTER(LHASH_NODE))),
    ('comp', LHASH_COMP_FN_TYPE),
    ('hash', LHASH_HASH_FN_TYPE),
    ('num_nodes', c_uint),
    ('num_alloc_nodes', c_uint),
    ('p', c_uint),
    ('pmax', c_uint),
    ('up_load', c_ulong),
    ('down_load', c_ulong),
    ('num_items', c_ulong),
    ('num_expands', c_ulong),
    ('num_expand_reallocs', c_ulong),
    ('num_contracts', c_ulong),
    ('num_contract_reallocs', c_ulong),
    ('num_hash_calls', c_ulong),
    ('num_comp_calls', c_ulong),
    ('num_insert', c_ulong),
    ('num_replace', c_ulong),
    ('num_delete', c_ulong),
    ('num_no_delete', c_ulong),
    ('num_retrieve', c_ulong),
    ('num_retrieve_miss', c_ulong),
    ('num_hash_comps', c_ulong),
    ('error', c_int),
]
x509_hash_dir_st._fields_ = [
    ('num_dirs', c_int),
    ('dirs', POINTER(STRING)),
    ('dirs_type', POINTER(c_int)),
    ('num_dirs_alloced', c_int),
]
x509_file_st._fields_ = [
    ('num_paths', c_int),
    ('num_alloced', c_int),
    ('paths', POINTER(STRING)),
    ('path_type', POINTER(c_int)),
]
class N14x509_object_st4DOT_36E(Union):
    pass
N14x509_object_st4DOT_36E._fields_ = [
    ('ptr', STRING),
    ('x509', POINTER(X509)),
    ('crl', POINTER(X509_CRL)),
    ('pkey', POINTER(EVP_PKEY)),
]
x509_object_st._fields_ = [
    ('type', c_int),
    ('data', N14x509_object_st4DOT_36E),
]
x509_lookup_method_st._fields_ = [
    ('name', STRING),
    ('new_item', CFUNCTYPE(c_int, POINTER(X509_LOOKUP))),
    ('free', CFUNCTYPE(None, POINTER(X509_LOOKUP))),
    ('init', CFUNCTYPE(c_int, POINTER(X509_LOOKUP))),
    ('shutdown', CFUNCTYPE(c_int, POINTER(X509_LOOKUP))),
    ('ctrl', CFUNCTYPE(c_int, POINTER(X509_LOOKUP), c_int, STRING, c_long, POINTER(STRING))),
    ('get_by_subject', CFUNCTYPE(c_int, POINTER(X509_LOOKUP), c_int, POINTER(X509_NAME), POINTER(X509_OBJECT))),
    ('get_by_issuer_serial', CFUNCTYPE(c_int, POINTER(X509_LOOKUP), c_int, POINTER(X509_NAME), POINTER(ASN1_INTEGER), POINTER(X509_OBJECT))),
    ('get_by_fingerprint', CFUNCTYPE(c_int, POINTER(X509_LOOKUP), c_int, POINTER(c_ubyte), c_int, POINTER(X509_OBJECT))),
    ('get_by_alias', CFUNCTYPE(c_int, POINTER(X509_LOOKUP), c_int, STRING, c_int, POINTER(X509_OBJECT))),
]
X509_VERIFY_PARAM_st._fields_ = [
    ('name', STRING),
    ('check_time', time_t),
    ('inh_flags', c_ulong),
    ('flags', c_ulong),
    ('purpose', c_int),
    ('trust', c_int),
    ('depth', c_int),
    ('policies', POINTER(STACK)),
]
x509_store_st._fields_ = [
    ('cache', c_int),
    ('objs', POINTER(STACK)),
    ('get_cert_methods', POINTER(STACK)),
    ('param', POINTER(X509_VERIFY_PARAM)),
    ('verify', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('verify_cb', CFUNCTYPE(c_int, c_int, POINTER(X509_STORE_CTX))),
    ('get_issuer', CFUNCTYPE(c_int, POINTER(POINTER(X509)), POINTER(X509_STORE_CTX), POINTER(X509))),
    ('check_issued', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(X509), POINTER(X509))),
    ('check_revocation', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('get_crl', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(POINTER(X509_CRL)), POINTER(X509))),
    ('check_crl', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(X509_CRL))),
    ('cert_crl', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(X509_CRL), POINTER(X509))),
    ('cleanup', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('ex_data', CRYPTO_EX_DATA),
    ('references', c_int),
]
x509_lookup_st._fields_ = [
    ('init', c_int),
    ('skip', c_int),
    ('method', POINTER(X509_LOOKUP_METHOD)),
    ('method_data', STRING),
    ('store_ctx', POINTER(X509_STORE)),
]
x509_store_ctx_st._fields_ = [
    ('ctx', POINTER(X509_STORE)),
    ('current_method', c_int),
    ('cert', POINTER(X509)),
    ('untrusted', POINTER(STACK)),
    ('crls', POINTER(STACK)),
    ('param', POINTER(X509_VERIFY_PARAM)),
    ('other_ctx', c_void_p),
    ('verify', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('verify_cb', CFUNCTYPE(c_int, c_int, POINTER(X509_STORE_CTX))),
    ('get_issuer', CFUNCTYPE(c_int, POINTER(POINTER(X509)), POINTER(X509_STORE_CTX), POINTER(X509))),
    ('check_issued', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(X509), POINTER(X509))),
    ('check_revocation', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('get_crl', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(POINTER(X509_CRL)), POINTER(X509))),
    ('check_crl', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(X509_CRL))),
    ('cert_crl', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX), POINTER(X509_CRL), POINTER(X509))),
    ('check_policy', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('cleanup', CFUNCTYPE(c_int, POINTER(X509_STORE_CTX))),
    ('valid', c_int),
    ('last_untrusted', c_int),
    ('chain', POINTER(STACK)),
    ('tree', POINTER(X509_POLICY_TREE)),
    ('explicit_policy', c_int),
    ('error_depth', c_int),
    ('error', c_int),
    ('current_cert', POINTER(X509)),
    ('current_issuer', POINTER(X509)),
    ('current_crl', POINTER(X509_CRL)),
    ('ex_data', CRYPTO_EX_DATA),
]
pkcs7_issuer_and_serial_st._fields_ = [
    ('issuer', POINTER(X509_NAME)),
    ('serial', POINTER(ASN1_INTEGER)),
]
pkcs7_signer_info_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('issuer_and_serial', POINTER(PKCS7_ISSUER_AND_SERIAL)),
    ('digest_alg', POINTER(X509_ALGOR)),
    ('auth_attr', POINTER(STACK)),
    ('digest_enc_alg', POINTER(X509_ALGOR)),
    ('enc_digest', POINTER(ASN1_OCTET_STRING)),
    ('unauth_attr', POINTER(STACK)),
    ('pkey', POINTER(EVP_PKEY)),
]
pkcs7_recip_info_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('issuer_and_serial', POINTER(PKCS7_ISSUER_AND_SERIAL)),
    ('key_enc_algor', POINTER(X509_ALGOR)),
    ('enc_key', POINTER(ASN1_OCTET_STRING)),
    ('cert', POINTER(X509)),
]
pkcs7_signed_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('md_algs', POINTER(STACK)),
    ('cert', POINTER(STACK)),
    ('crl', POINTER(STACK)),
    ('signer_info', POINTER(STACK)),
    ('contents', POINTER(pkcs7_st)),
]
pkcs7_enc_content_st._fields_ = [
    ('content_type', POINTER(ASN1_OBJECT)),
    ('algorithm', POINTER(X509_ALGOR)),
    ('enc_data', POINTER(ASN1_OCTET_STRING)),
    ('cipher', POINTER(EVP_CIPHER)),
]
pkcs7_enveloped_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('recipientinfo', POINTER(STACK)),
    ('enc_data', POINTER(PKCS7_ENC_CONTENT)),
]
pkcs7_signedandenveloped_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('md_algs', POINTER(STACK)),
    ('cert', POINTER(STACK)),
    ('crl', POINTER(STACK)),
    ('signer_info', POINTER(STACK)),
    ('enc_data', POINTER(PKCS7_ENC_CONTENT)),
    ('recipientinfo', POINTER(STACK)),
]
pkcs7_digest_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('md', POINTER(X509_ALGOR)),
    ('contents', POINTER(pkcs7_st)),
    ('digest', POINTER(ASN1_OCTET_STRING)),
]
pkcs7_encrypted_st._fields_ = [
    ('version', POINTER(ASN1_INTEGER)),
    ('enc_data', POINTER(PKCS7_ENC_CONTENT)),
]
class N8pkcs7_st4DOT_37E(Union):
    pass
N8pkcs7_st4DOT_37E._fields_ = [
    ('ptr', STRING),
    ('data', POINTER(ASN1_OCTET_STRING)),
    ('sign', POINTER(PKCS7_SIGNED)),
    ('enveloped', POINTER(PKCS7_ENVELOPE)),
    ('signed_and_enveloped', POINTER(PKCS7_SIGN_ENVELOPE)),
    ('digest', POINTER(PKCS7_DIGEST)),
    ('encrypted', POINTER(PKCS7_ENCRYPT)),
    ('other', POINTER(ASN1_TYPE)),
]
pkcs7_st._fields_ = [
    ('asn1', POINTER(c_ubyte)),
    ('length', c_long),
    ('state', c_int),
    ('detached', c_int),
    ('type', POINTER(ASN1_OBJECT)),
    ('d', N8pkcs7_st4DOT_37E),
]
class N15STORE_OBJECT_st4DOT_38E(Union):
    pass
class N15STORE_OBJECT_st4DOT_384DOT_39E(Structure):
    pass
N15STORE_OBJECT_st4DOT_384DOT_39E._fields_ = [
    ('status', STORE_CERTIFICATE_STATUS),
    ('certificate', POINTER(X509)),
]
N15STORE_OBJECT_st4DOT_38E._fields_ = [
    ('x509', N15STORE_OBJECT_st4DOT_384DOT_39E),
    ('crl', POINTER(X509_CRL)),
    ('key', POINTER(EVP_PKEY)),
    ('number', POINTER(BIGNUM)),
    ('arbitrary', POINTER(BUF_MEM)),
]
STORE_OBJECT_st._fields_ = [
    ('type', STORE_OBJECT_TYPES),
    ('data', N15STORE_OBJECT_st4DOT_38E),
]
STORE_attr_info_st._fields_ = [
]
err_state_st._fields_ = [
    ('pid', c_ulong),
    ('err_flags', c_int * 16),
    ('err_buffer', c_ulong * 16),
    ('err_data', STRING * 16),
    ('err_data_flags', c_int * 16),
    ('err_file', STRING * 16),
    ('err_line', c_int * 16),
    ('top', c_int),
    ('bottom', c_int),
]
ERR_string_data_st._fields_ = [
    ('error', c_ulong),
    ('string', STRING),
]
ENGINE_CMD_DEFN_st._fields_ = [
    ('cmd_num', c_uint),
    ('cmd_name', STRING),
    ('cmd_desc', STRING),
    ('cmd_flags', c_uint),
]
st_dynamic_MEM_fns._fields_ = [
    ('malloc_cb', dyn_MEM_malloc_cb),
    ('realloc_cb', dyn_MEM_realloc_cb),
    ('free_cb', dyn_MEM_free_cb),
]
st_dynamic_LOCK_fns._fields_ = [
    ('lock_locking_cb', dyn_lock_locking_cb),
    ('lock_add_lock_cb', dyn_lock_add_lock_cb),
    ('dynlock_create_cb', dyn_dynlock_create_cb),
    ('dynlock_lock_cb', dyn_dynlock_lock_cb),
    ('dynlock_destroy_cb', dyn_dynlock_destroy_cb),
]
st_dynamic_fns._fields_ = [
    ('static_state', c_void_p),
    ('err_fns', POINTER(ERR_FNS)),
    ('ex_data_fns', POINTER(CRYPTO_EX_DATA_IMPL)),
    ('mem_fns', dynamic_MEM_fns),
    ('lock_fns', dynamic_LOCK_fns),
]
_IO_marker._fields_ = [
    ('_next', POINTER(_IO_marker)),
    ('_sbuf', POINTER(_IO_FILE)),
    ('_pos', c_int),
]
__locale_data._fields_ = [
]
AUTHORITY_KEYID_st._fields_ = [
]
__all__ = ['__uint16_t', 'pkcs7_enc_content_st', '__int16_t',
           'X509_REVOKED', 'STORE_START_OBJECT_FUNC_PTR',
           'st_dynamic_LOCK_fns', 'SSL_CTX', 'openssl_item_st',
           '__off64_t', 'STORE_END_OBJECT_FUNC_PTR', 'fpos_t', 'X509',
           'STORE_GENERIC_FUNC_PTR', 'dyn_dynlock_create_cb',
           '_G_int16_t', 'STORE_X509_EXPIRED', '__FILE',
           'ASN1_TEMPLATE', 'BIO_METHOD', '__time_t', 'dh_method',
           'bio_f_buffer_ctx_struct', 'X509_SIG', 'X509_algor_st',
           'STORE_GET_OBJECT_FUNC_PTR', 'X509_POLICY_NODE_st',
           'pid_t', 'ec_group_st', 'ASN1_IA5STRING',
           'dynamic_v_check_fn', '__uint64_t', 'x509_cinf_st',
           'buf_mem_st', 'STORE_ATTR_SUBJECTKEYID',
           'ecpk_parameters_st', 'BIGNUM', 'PBEPARAM',
           'X509_NAME_ENTRY', '__clockid_t', 'lhash_st', 'id_t',
           'dyn_MEM_malloc_cb', 'ASN1_BIT_STRING', '_G_fpos_t',
           'STORE_METHOD', 'STORE_X509_REVOKED', 'STORE_ATTR_END',
           'env_md_ctx_st', 'pkcs7_signer_info_st', 'ASN1_METHOD',
           '__locale_data', '__u_long', 'DSA_SIG', 'DSA',
           'STORE_CERTIFICATE_STATUS', 'NETSCAPE_CERT_SEQUENCE',
           'pthread_t', 'UI_STRING', '__io_read_fn',
           'ENGINE_GEN_FUNC_PTR', 'DES_cblock', '__mode_t',
           'STORE_MODIFY_OBJECT_FUNC_PTR', '__off_t',
           'STORE_PARAM_KEY_PARAMETERS', '_ossl_old_des_ks_struct',
           'SHAstate_st', 'u_quad_t', 'STORE_OBJECT_TYPE_NUM',
           'fsfilcnt64_t', 'daddr_t', 'ui_string_st', 'x509_file_st',
           'X509_req_info_st', 'evp_Encode_Ctx_st', 'engine_st',
           'CRYPTO_EX_DATA', '__int8_t', 'HMAC_CTX', '__fsblkcnt64_t',
           'ec_point_st', 'X509_POLICY_LEVEL', 'EC_KEY',
           'EVP_CIPHER_CTX', 'bn_gencb_st', 'EC_POINT', 'RSA_METHOD',
           'timer_t', '__fsfilcnt64_t', 'dyn_MEM_free_cb', '_IO_FILE',
           'pthread_key_t', 'STORE_ATTR_ISSUERKEYID', 'ui_st',
           'X509_PUBKEY', '__locale_struct', 'u_int8_t',
           'ASN1_ITEM_st', 'pkcs7_recip_info_st', 'off_t',
           'ui_method_st', 'crypto_ex_data_st', '__fsblkcnt_t',
           '__locale_t', 'ENGINE_LOAD_KEY_PTR', 'ECDH_METHOD',
           'CRYPTO_EX_dup', 'OCSP_RESPID', 'BN_MONT_CTX',
           'STORE_ATTR_FRIENDLYNAME', 'STORE_PARAM_EVP_TYPE',
           'ASN1_NULL', 'ASN1_INTEGER', 'CRYPTO_EX_new',
           'SHA256state_st', 'asn1_type_st', 'CRYPTO_EX_DATA_FUNCS',
           'key_t', 'uint', 'ASN1_VALUE_st', 'DH_METHOD',
           'STORE_OBJECT_TYPE_NUMBER', '__pthread_internal_slist',
           'RSA', '__u_int', 'asn1_output_data_fn', 'ssize_t',
           '__clock_t', 'fsblkcnt_t', 'X509_PKEY',
           'evp_verify_method', 'FILE', 'X509_POLICY_CACHE_st',
           'X509_sig_st', 'SHA512_CTX', 'pkcs7_issuer_and_serial_st',
           'CRYPTO_MEM_LEAK_CB', 'X509_NAME', 'blkcnt_t',
           'evp_cipher_info_st', 'BN_BLINDING', '__qaddr_t',
           'err_state_st', 'st_dynamic_MEM_fns', 'DES_key_schedule',
           'N11__mbstate_t3DOT_2E', 'u_char', 'fpos64_t', 'uid_t',
           'cookie_write_function_t', 'u_int64_t', 'ASN1_STRING',
           'N11evp_pkey_st4DOT_28E', 'STORE_OBJECT_TYPE_X509_CRL',
           'sigset_t', 'conf_st', 'ASN1_CTX',
           'N12asn1_type_st4DOT_27E', 'PKCS7', '__int32_t',
           'UI_METHOD', 'NETSCAPE_SPKI', 'st_CRYPTO_EX_DATA_IMPL',
           'cast_key_st', 'X509_HASH_DIR_CTX', 'ERR_string_data_st',
           'st_dynamic_fns', '__fd_mask', 'dynamic_MEM_fns',
           'STORE_PARAM_TYPE_NUM', 'clock_t', 'aes_key_st',
           'asn1_string_table_st', '__useconds_t',
           'X509_POLICY_TREE_st', 'ASN1_VISIBLESTRING',
           'dynamic_LOCK_fns', 'x509_cert_pair_st', 'DSA_SIG_st',
           'obj_name_st', 'X509_LOOKUP_METHOD', 'u_int32_t',
           '_IO_cookie_io_functions_t', 'EVP_CIPHER_INFO',
           '__gnuc_va_list', 'AES_KEY', 'PKCS7_ISSUER_AND_SERIAL',
           'BN_CTX', 'ERR_FNS', 'SHA_CTX', 'pkcs7_signed_st', 'SSL',
           'EVP_MD', 'ASN1_BOOLEAN', '__rlim64_t', 'ino_t',
           'ecdsa_method', 'dyn_dynlock_lock_cb', 'ASN1_OCTET_STRING',
           'STORE_PARAM_AUTH_KRB5_TICKET', '__blksize_t',
           'ENGINE_CMD_DEFN', 'pthread_spinlock_t', 'asn1_ctx_st',
           '__pthread_slist_t', 'ecdh_method', 'BIO_F_BUFFER_CTX',
           'ECDSA_SIG_st', 'bn_mont_ctx_st', 'X509_REQ_INFO',
           'OCSP_REQ_CTX', 'ERR_STATE', 'STORE_certificate_status',
           'STORE_CTRL_FUNC_PTR', 'STORE_ATTR_SERIAL', 'ino64_t',
           'x509_attributes_st', 'ec_method_st', '__mbstate_t',
           'asn1_object_st', 'ASN1_ENCODING', '__uint8_t',
           'LHASH_NODE', '__u_char', '__caddr_t', '__blkcnt64_t',
           'PKCS7_SIGNER_INFO', 'asn1_method_st',
           'STORE_GENERATE_OBJECT_FUNC_PTR', 'stack_st',
           'bio_info_cb', 'STORE_ATTR_SUBJECT',
           'N18x509_attributes_st4DOT_35E', 'ASN1_PRINTABLESTRING',
           'PBEPARAM_st', 'quad_t', 'STORE_attr_info_st', 'rsa_st',
           'ASN1_UNIVERSALSTRING', 'X509_OBJECT', 'DH',
           'bn_blinding_st', 'ASN1_ENCODING_st', 'ASN1_TLC_st',
           '__rlim_t', 'BIO', 'nlink_t', 'BUF_MEM', 'bio_method_st',
           'BIO_dummy', 'ssl_ctx_st', 'RC4_KEY',
           'BIT_STRING_BITNAME_st', 'ulong', 'STORE_ATTR_TYPES',
           'EVP_ENCODE_CTX', 'STORE_ATTR_ISSUER', 'STORE_params',
           'int8_t', 'OBJ_NAME', 'STORE_PARAM_BITS',
           'ENGINE_CIPHERS_PTR', 'STORE', 'X509_POLICY_CACHE',
           'PKCS8_PRIV_KEY_INFO', 'X509_CRL', '__fsfilcnt_t',
           '__quad_t', '__key_t', 'X509_VAL', 'dev_t',
           'ASN1_TEMPLATE_st', 'ENGINE', 'LHASH_DOALL_ARG_FN_TYPE',
           'ocsp_req_ctx_st', 'ASN1_ITEM_EXP', 'fsfilcnt_t',
           '__swblk_t', 'mode_t', 'ASN1_VALUE', 'X509_POLICY_NODE',
           'EVP_PKEY', 'CRYPTO_EX_free', 'ENGINE_SSL_CLIENT_CERT_PTR',
           '_ossl_old_des_cblock', 'X509_INFO', 'asn1_string_st',
           '__loff_t', 'RAND_METHOD', 'STORE_ATTR_FILENAME',
           'ASN1_const_CTX', 'env_md_st', 'cookie_seek_function_t',
           'STORE_NEXT_OBJECT_FUNC_PTR', 'LHASH',
           'PKCS7_SIGN_ENVELOPE', 'X509_extension_st', 'va_list',
           'PKCS7_DIGEST', 'N6DES_ks4DOT_30E', 'fd_mask', 'ASN1_TYPE',
           'PKCS7_SIGNED', '__fsid_t', 'cookie_close_function_t',
           'EC_METHOD', '__ssize_t', 'comparison_fn_t', 'BF_KEY',
           'hmac_ctx_st', 'SHA512state_st', 'OPENSSL_ITEM', 'size_t',
           'int16_t', 'BN_GENCB', 'STORE_OBJECT_TYPE_PRIVATE_KEY',
           '__sigset_t', 'X509_TRUST', 'X509_STORE',
           'STORE_PARAM_AUTH_PASSPHRASE', 'X509_STORE_CTX',
           'ASN1_TIME', 'STORE_ATTR_KEYID', 'X509_LOOKUP',
           'Netscape_spki_st', 'STORE_X509_VALID',
           'ENGINE_CTRL_FUNC_PTR', '__intptr_t', 'SHA256_CTX',
           'STORE_CLEANUP_FUNC_PTR', 'x509_cert_aux_st', 'ushort',
           'STORE_STORE_OBJECT_FUNC_PTR', '__blkcnt_t', 'clockid_t',
           'N8pkcs7_st4DOT_37E', 'BN_RECP_CTX', 'x509_lookup_st',
           'ASN1_BMPSTRING', 'asn1_header_st', 'x509_trust_st',
           'STORE_OBJECT_TYPES', 'int32_t', 'off64_t',
           'X509_CRL_INFO', 'st_ERR_FNS', 'dynamic_fns',
           'ASN1_HEADER', 'X509_crl_info_st', 'i2d_of_void',
           'LHASH_HASH_FN_TYPE', 'N15STORE_OBJECT_st4DOT_38E',
           '__compar_d_fn_t', 'ssl_st', 'ENGINE_GEN_INT_FUNC_PTR',
           'evp_pkey_st', 'pkcs7_signedandenveloped_st', 'X509V3_CTX',
           'STORE_INITIALISE_FUNC_PTR', 'EVP_PBE_KEYGEN', '__dev_t',
           'ASN1_UTCTIME', 'STORE_ATTR_TYPE_NUM',
           'dynamic_bind_engine', 'X509_POLICY_LEVEL_st',
           'crypto_ex_data_func_st', '__suseconds_t', 'STORE_ATTR_OR',
           'u_long', 'PBKDF2PARAM_st', 'ECDSA_SIG', 'rc4_key_st',
           'DSA_METHOD', 'STORE_OBJECT', 'EVP_CIPHER',
           'ocsp_response_st', 'BIT_STRING_BITNAME',
           'PKCS7_RECIP_INFO', 'EC_GROUP', 'X509_revoked_st',
           'X509_CERT_AUX', 'X509_VERIFY_PARAM',
           'X509_VERIFY_PARAM_st', 'STORE_OBJECT_TYPE_PUBLIC_KEY',
           'time_t', 'u_short', 'N23_ossl_old_des_ks_struct4DOT_31E',
           'error_t', '_ossl_old_des_key_schedule', '_IO_marker',
           'fsblkcnt64_t', 'v3_ext_ctx', 'ASN1_GENERALSTRING',
           'STACK', 'X509_name_entry_st', '__io_write_fn', 'caddr_t',
           '__ino_t', 'bignum_st', 'X509_CINF', '_IO_lock_t',
           'ASN1_TLC', 'PKCS7_ENCRYPT', 'NETSCAPE_SPKAC',
           'Netscape_spkac_st', 'X509_CERT_PAIR',
           'dyn_lock_add_lock_cb', '__int64_t', 'pkcs7_st',
           'ocsp_responder_id_st', 'rand_meth_st', 'suseconds_t',
           'ASN1_OBJECT', 'X509_val_st', 'private_key_st',
           'X509_objects_st', 'CRYPTO_EX_DATA_IMPL', 'pthread_once_t',
           '__timer_t', 'CONF', 'ERR_STRING_DATA', 'u_int16_t',
           '__uint32_t', 'register_t', 'rsa_meth_st', 'X509_crl_st',
           '__ino64_t', 'N14SHA512state_st4DOT_34E',
           'store_method_st', 'loff_t', 'PBE2PARAM', 'blksize_t',
           'STORE_attribs', 'Netscape_certificate_sequence',
           'bignum_ctx', '_G_uint32_t', 'ASN1_UTF8STRING',
           'dyn_MEM_realloc_cb', 'pkcs7_encrypted_st',
           'STORE_ATTR_ISSUERSERIALHASH', '_G_fpos64_t',
           'STORE_X509_SUSPENDED', 'PBKDF2PARAM', '__nlink_t',
           '__compar_fn_t', 'CAST_KEY', '__io_close_fn',
           'LHASH_DOALL_FN_TYPE', 'OCSP_RESPONSE',
           'evp_cipher_ctx_st', 'X509_ALGOR', '__id_t',
           'cookie_io_functions_t', 'const_DES_cblock',
           'dyn_dynlock_destroy_cb', 'dsa_st', 'dyn_lock_locking_cb',
           'STORE_PARAM_KEY_NO_PARAMETERS', 'X509_EXTENSION',
           'PKCS7_ENVELOPE', 'N11bn_gencb_st4DOT_26E', 'ec_key_st',
           'x509_lookup_method_st', '_G_uint16_t',
           'ENGINE_DIGESTS_PTR', 'STORE_OBJECT_TYPE_ARBITRARY',
           'bn_recp_ctx_st', 'X509_info_st', 'x509_store_st',
           'X509_pubkey_st', 'pkcs7_digest_st', 'ASN1_STRING_TABLE',
           'ENGINE_CMD_DEFN_st', 'X509_EXTENSIONS',
           'pkcs7_enveloped_st', 'UI', 'ptrdiff_t', 'X509_REQ',
           'CRYPTO_dynlock_value', 'X509_req_st', 'x509_store_ctx_st',
           'asn1_const_ctx_st', 'lhash_node_st', 'store_st',
           'LHASH_COMP_FN_TYPE', 'STORE_OBJECT_st', '__gid_t',
           'STORE_OBJECT_TYPE_X509_CERTIFICATE', 'd2i_of_void',
           'X509_ATTRIBUTE', 'ECDSA_METHOD', '__daddr_t',
           '__sig_atomic_t', '__io_seek_fn', 'u_int',
           'ASN1_T61STRING', 'gid_t', 'ASN1_ITEM', 'ASN1_ENUMERATED',
           'pkcs8_priv_key_info_st', 'PBE2PARAM_st',
           'N15STORE_OBJECT_st4DOT_384DOT_39E',
           'cookie_read_function_t', 'blkcnt64_t', 'X509_OBJECTS',
           'x509_object_st', 'STORE_object_types', 'DES_ks',
           'N14x509_object_st4DOT_36E', 'dsa_method', 'int64_t',
           'bio_st', 'bf_key_st', 'ASN1_GENERALIZEDTIME',
           'PKCS7_ENC_CONTENT', 'ECPKPARAMETERS', 'EVP_MD_CTX',
           '__u_quad_t', '__u_short', 'evp_cipher_st', 'X509_name_st',
           'x509_st', 'fsid_t', '__pid_t', 'AUTHORITY_KEYID_st',
           'STORE_PARAM_TYPES', 'x509_hash_dir_st', 'evp_sign_method',
           'X509_POLICY_TREE', 'useconds_t',
           'STORE_HANDLE_OBJECT_FUNC_PTR', '__uid_t', 'locale_t',
           'STORE_ATTR_CERTHASH', 'X509_ALGORS', '__socklen_t',
           'dh_st', 'STORE_ATTR_INFO', '_G_int32_t',
           'STORE_ATTR_EMAIL', 'X509_CERT_FILE_CTX']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging, re

log=logging.getLogger('preprocess')

class HeaderCleaner:
  '''
    Cleans a Preprocessed File for gccxml comsumption.
    Strips off static functions and extern references.
    
    @param preprocessed: the preprocessed file name
    @param out: the desired output file
  '''
  functionsReOneLine = r"""  # static inline bool kernel_page_present(struct page *page) { return true; } 
  ^ ((static\ (inline|__inline__)) 
         (\s+__attribute__\(\(always_inline\)\))*  (?P<sig> \s+\w+)* (\s*[*]\s*)* 
                (?P<funcname>  \w+ ) (?P<args> \s*\([^{;]+?\)\s* ) 
        ( { }$ |{ .*? }$  )
    )     
   """
  functionsRe = r"""  # nice - ok for pointers
  ^ ((static\ (inline|__inline__)) 
         (\s+__attribute__\(\(always_inline\)\))*  (?P<sig> \s+\w+)* (\s*[*]\s*)* 
                (?P<funcname>  \w+ ) (?P<args> \s*\([^{;]+?\)\s* ) 
        ( { }$ |{ . }$ | {  .*?  ^}$  )
    )     
   """
  functionsRe2 = r"""  # nice - ok for pointers
  ^ ( (__extension__\s+)*   (extern\ (inline|__inline__|__inline)) 
         (\s+__attribute__\(\((always_inline|__gnu_inline__)\)\))*  (?P<sig> \s+\w+)* (\s*[*]\s*)* 
                (?P<funcname>  \w+ ) (?P<args> \s*\([^{;]+?\)\s* ) 
        ( { }$ |{ . }$ | {  .*?  ^}$  )
    )     
   """
  externsRe = r"""  # extern functions sig.
  ^((extern) \s+ (?!struct|enum) [^{]*? ;$  ) 
   """
  externsRe2 = r"""  # functions sig. on one line
  ^((void|unsigned|long|int) \s+ (?!struct|enum) .*? ;$  ) 
   """
  def __init__(self, preprocessed, out):
    self.preprocessed = file(preprocessed).read()
    self.out = out
    
  def stripFunctions(self, data):
    # delete oneliner
    REGEX_OBJ = re.compile( self.functionsReOneLine, re.MULTILINE| re.VERBOSE)
    data1 = REGEX_OBJ.sub(r'/** // supprimed function a:  */',data)

    REGEX_OBJ = re.compile( self.functionsRe, re.MULTILINE| re.VERBOSE | re.DOTALL)
    data2 = REGEX_OBJ.sub(r'/** // supprimed function b:  */',data1)

    REGEX_OBJ = re.compile( self.functionsRe2, re.MULTILINE| re.VERBOSE | re.DOTALL)
    data3 = REGEX_OBJ.sub(r'/** // supprimed function c:  */',data2)
    return data3


  def stripExterns(self, data):
    REGEX_OBJ2 = re.compile( self.externsRe, re.MULTILINE| re.VERBOSE | re.DOTALL)
    data1 = REGEX_OBJ2.sub(r'/** // supprimed extern  */', data)

    REGEX_OBJ2 = re.compile( self.externsRe2, re.MULTILINE| re.VERBOSE | re.DOTALL)
    data2 = REGEX_OBJ2.sub(r'/** // supprimed function sig  */', data1)
    return data2

  def changeReservedWords(self, data):
    data1 = data.replace('*new);','*new1);')
    data1 = data1.replace('*proc_handler;','*proc_handler1;')
    mre = re.compile(r'\bprivate;')
    data2 = mre.sub('private1;',data1)
    mre = re.compile(r'\bnamespace\b')
    data3 = mre.sub('namespace1',data2)
    return data3

  def clean(self):
    data2 = self.stripFunctions(self.preprocessed)
    data3 = self.stripExterns(data2)
    #data3 = data2
    data4 = self.changeReservedWords(data3)
    self.fout = file(self.out,'w')
    return self.fout.write(data4)



def clean(prepro, out):
  clean = HeaderCleaner(prepro, out)
  return clean.clean()

#clean('ctypes_linux_generated.c','ctypes_linux_generated_clean.c')


data='''
typedef int (*pte_fn_t)(pte_t *pte, pgtable_t token, unsigned long addr,
   void *data);
extern int apply_to_page_range(struct mm_struct *mm, unsigned long address,
          unsigned long size, pte_fn_t fn, void *data);
void vm_stat_account(struct mm_struct *, unsigned long, struct file *, long);
static inline void
kernel_map_pages(struct page *page, int numpages, int enable) {}
static inline void enable_debug_pagealloc(void)
{
}
static inline bool kernel_page_present(struct page *page) { return true; }
extern struct vm_area_struct *get_gate_vma(struct task_struct *tsk);
int in_gate_area_no_task(unsigned long addr);
int in_gate_area(struct task_struct *task, unsigned long addr);
'''

from cleaner import HeaderCleaner
import re
def testF(data):
  obj = re.compile( HeaderCleaner.functionsRe, re.MULTILINE| re.VERBOSE | re.DOTALL)
  for p in obj.findall(data):
    print p[0]
    print '----------------------'
  return

def testF2(data):
  obj = re.compile( HeaderCleaner.functionsRe2, re.MULTILINE| re.VERBOSE | re.DOTALL)
  for p in obj.findall(data):
    print p[0]
    print '----------------------'
  return

def testE(data):
  obj = re.compile( HeaderCleaner.externsRe, re.MULTILINE| re.VERBOSE | re.DOTALL)
  for p in obj.findall(data):
    print p[0]
    print '----------------------'
  return

def testE2(data):
  obj = re.compile( HeaderCleaner.externsRe2, re.MULTILINE| re.VERBOSE | re.DOTALL)
  for p in obj.findall(data):
    print p[0]
    print '----------------------'
  return

'''
HeaderCleaner.externsRe = r"""  # extern functions sig.
  ^((extern) \s+ (?!struct|enum) .*? ;$  ) 
   """
HeaderCleaner.externsRe2 = r"""  # functions sig. on one line
  ^((void|unisgned|long|int) \s+ (?!struct|enum) .*? ;$  ) 
   """


c = HeaderCleaner('/dev/null','/dev/null')
data1 = c.stripFunctions( data )
data2 = c.stripExterns( data1 )

print data1


'mf_flags' in data
'mf_flags' in data1
'mf_flags' in data2


testE(data)


'''











#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import ctypes
import logging, sys

''' insure ctypes basic types are subverted '''
from haystack import model

from haystack.model import is_valid_address,is_valid_address_value,getaddress,array2bytes,bytes2array
from haystack.model import LoadableMembers,RangeValue,NotNull,CString

import ctypes_openssl_generated as gen

log=logging.getLogger('openssl.model')

''' hmac.h:69 '''
HMAC_MAX_MD_CBLOCK=128
''' evp.h:91 '''
EVP_MAX_BLOCK_LENGTH=32
EVP_MAX_IV_LENGTH=16
AES_MAXNR=14 # aes.h:66
RIJNDAEL_MAXNR=14


# ============== Internal type defs ==============


class OpenSSLStruct(LoadableMembers):
  ''' defines classRef '''
  pass

BN_ULONG = ctypes.c_ulong

# evp/e_aes.c:66
class EVP_AES_KEY(OpenSSLStruct):
  _fields_ = [
  ('ks', gen.AES_KEY),
	]
  def fromPyObj(self,pyobj):
    self.ks = gen.AES_KEY().fromPyObj(pyobj.ks)
    return self

class EVP_RC4_KEY(OpenSSLStruct): # evp/e_rca.c
  _fields_ = [
  ('ks', gen.RC4_KEY)
  ]




################ START copy generated classes ##########################

# copy generated classes (gen.*) to this module as wrapper
model.copyGeneratedClasses(gen, sys.modules[__name__])

# register all classes (gen.*, locally defines, and local duplicates) to haystack
# create plain old python object from ctypes.Structure's, to picke them
model.registerModule(sys.modules[__name__])

################ END   copy generated classes ##########################




############# Start expectedValues and methods overrides #################



''' aes.h:78 '''
####### AES_KEY #######
def AES_KEY_getKey(self):
  #return array2bytes(self.rd_key)
  return ','.join(["0x%lx"%key for key in self.rd_key])
def AES_KEY_getRounds(self):
  return self.rounds
def AES_KEY_fromPyObj(self,pyobj):
  #copy rd_key
  self.rd_key=bytes2array(pyobj.rd_key,ctypes.c_ulong)
  #copy rounds
  self.rounds=pyobj.rounds
  return self
AES_KEY.getKey = AES_KEY_getKey
AES_KEY.getRounds = AES_KEY_getRounds
AES_KEY.fromPyObj = AES_KEY_fromPyObj
#######


############ BIGNUM
BIGNUM.expectedValues={
    "neg": [0,1]
  }
def BIGNUM_loadMembers(self, mappings, maxDepth):
  ''' 
  #self._d = process.readArray(attr_obj_address, ctypes.c_ulong, self.top) 
  ## or    
  #ulong_array= (ctypes.c_ulong * self.top)    
  '''
  if not self.isValid(mappings):
    log.debug('BigNUm tries to load members when its not validated')
    return False
  # Load and memcopy d / BN_ULONG *
  attr_obj_address=getaddress(self.d)
  if not bool(self.d):
    log.debug('BIGNUM has a Null pointer d')
    return True
  memoryMap = is_valid_address_value( attr_obj_address, mappings)
  contents=(BN_ULONG*self.top).from_buffer_copy(memoryMap.readArray(attr_obj_address, BN_ULONG, self.top))
  log.debug('contents acquired %d'%ctypes.sizeof(contents))
  self.d.contents=BN_ULONG.from_address(ctypes.addressof(contents))
  self.d=ctypes.cast(contents, ctypes.POINTER(BN_ULONG) ) 
  return True

def BIGNUM_isValid(self,mappings):
  if ( self.dmax < 0 or self.top < 0 or self.dmax < self.top ):
    return False
  return LoadableMembers.isValid(self,mappings)

def BIGNUM___str__(self):
  d= getaddress(self.d)
  return ("BN { d=0x%lx, top=%d, dmax=%d, neg=%d, flags=%d }"%
              (d, self.top, self.dmax, self.neg, self.flags) )

BIGNUM.loadMembers = BIGNUM_loadMembers
BIGNUM.isValid     = BIGNUM_isValid
BIGNUM.__str__     = BIGNUM___str__
#################


# CRYPTO_EX_DATA crypto.h:158:
def CRYPTO_EX_DATA_loadMembers(self, mappings, maxDepth):
  ''' erase self.sk'''
  #self.sk=ctypes.POINTER(STACK)()
  return LoadableMembers.loadMembers(self, mappings, maxDepth)
def CRYPTO_EX_DATA_isValid(self,mappings):
  ''' erase self.sk'''
  # TODO why ?
  #self.sk=ctypes.POINTER(STACK)()
  return LoadableMembers.isValid(self,mappings)

CRYPTO_EX_DATA.loadMembers = CRYPTO_EX_DATA_loadMembers 
CRYPTO_EX_DATA.isValid     = CRYPTO_EX_DATA_isValid 
#################


######## RSA key
RSA.expectedValues={
    "pad": [0], 
    "version": [0], 
    "references": RangeValue(0,0xfff),
    "n": [NotNull],
    "e": [NotNull],
    "d": [NotNull],
    "p": [NotNull],
    "q": [NotNull],
    "dmp1": [NotNull],
    "dmq1": [NotNull],
    "iqmp": [NotNull]
  }
def RSA_printValid(self,mappings):
  log.debug( '----------------------- LOADED: %s'%self.loaded)
  log.debug('pad: %d version %d ref %d'%(self.pad,self.version,self.references) )
  log.debug(is_valid_address( self.n, mappings)    )
  log.debug(is_valid_address( self.e, mappings)    )
  log.debug(is_valid_address( self.d, mappings)    )
  log.debug(is_valid_address( self.p, mappings)    )
  log.debug(is_valid_address( self.q, mappings)    )
  log.debug(is_valid_address( self.dmp1, mappings) ) 
  log.debug(is_valid_address( self.dmq1, mappings) )
  log.debug(is_valid_address( self.iqmp, mappings) )
  return
def RSA_loadMembers(self, mappings, maxDepth):
  #self.meth = 0 # from_address(0)
  # ignore bignum_data.
  #self.bignum_data = 0
  self.bignum_data.ptr.value = 0
  #self.blinding = 0
  #self.mt_blinding = 0

  if not LoadableMembers.loadMembers(self, mappings, maxDepth):
    log.debug('RSA not loaded')
    return False
  return True

RSA.printValid  = RSA_printValid
RSA.loadMembers = RSA_loadMembers



########## DSA Key
DSA.expectedValues={
    "pad": [0], 
    "version": [0], 
    "references": RangeValue(0,0xfff),
    "p": [NotNull],
    "q": [NotNull],
    "g": [NotNull],
    "pub_key": [NotNull],
    "priv_key": [NotNull]
  }
def DSA_printValid(self,mappings):
  log.debug( '----------------------- \npad: %d version %d ref %d'%(self.pad,self.version,self.write_params) )
  log.debug(is_valid_address( self.p, mappings)    )
  log.debug(is_valid_address( self.q, mappings)    )
  log.debug(is_valid_address( self.g, mappings)    )
  log.debug(is_valid_address( self.pub_key, mappings)    )
  log.debug(is_valid_address( self.priv_key, mappings)    )
  return
def DSA_loadMembers(self, mappings, maxDepth):
  # clean other structs
  # r and kinv can be null
  self.meth = None
  self._method_mod_p = None
  #self.engine = None
  
  if not LoadableMembers.loadMembers(self, mappings, maxDepth):
    log.debug('DSA not loaded')
    return False

  return True

DSA.printValid  = DSA_printValid
DSA.loadMembers = DSA_loadMembers

######### EP_CIPHER
EVP_CIPHER.expectedValues={
    "key_len": RangeValue(1,0xff), # key_len *8 bits ..2040 bits for a key is enought ? 
                                   # Default value for variable length ciphers 
    "iv_len": RangeValue(1,0xff), #  
    "init": [NotNull], 
    "do_cipher": [NotNull], 
    #"cleanup": [NotNull], # aes-cbc ?
    "ctx_size": RangeValue(0,0xffff), #  app_data struct should not be too big
  }


########### EVP_CIPHER_CTX
EVP_CIPHER_CTX.expectedValues={
    "cipher": [NotNull], 
    "encrypt": [0,1], 
    "buf_len": RangeValue(0,EVP_MAX_BLOCK_LENGTH), ## number we have left, so must be less than buffer_size
    #"engine": , # can be null
    #"app_data": , # can be null if cipher_data is not
    #"cipher_data": , # can be null if app_data is not
    "key_len": RangeValue(1,0xff), # key_len *8 bits ..2040 bits for a key is enought ? 
  }
  
def EVP_CIPHER_CTX_getOIV(self):
  return array2bytes(self.oiv)
def EVP_CIPHER_CTX_getIV(self):
  return array2bytes(self.iv)

EVP_CIPHER_CTX.getOIV = EVP_CIPHER_CTX_getOIV
EVP_CIPHER_CTX.getIV  = EVP_CIPHER_CTX_getIV

##########


# checkks
'''
src=sys.modules[__name__]
for (name, klass) in inspect.getmembers(src, inspect.isclass):
  if klass.__module__ == src.__name__ or klass.__module__.endswith('%s_generated'%(src.__name__) ) :
    if not klass.__name__.endswith('_py'):
      print klass, len(klass.classRef)
'''

def printSizeof(mini=-1):
  for (name,klass) in inspect.getmembers(sys.modules[__name__], inspect.isclass):
    if type(klass) == type(ctypes.Structure) and klass.__module__.endswith('%s_generated'%(__name__) ) :
      if ctypes.sizeof(klass) > mini:
        print '%s:'%name,ctypes.sizeof(klass)
  #print 'SSLCipherSuiteInfo:',ctypes.sizeof(SSLCipherSuiteInfo)
  #print 'SSLChannelInfo:',ctypes.sizeof(SSLChannelInfo)


if __name__ == '__main__':
  printSizeof()




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import logging, os, subprocess, sys

log=logging.getLogger('generate')

class Generator:
  '''
    Wrapper around Ctypeslib
    Make python objects out of structures from C headers
    
    @param cleaned: the desired input c file name, preprocessed and cleaned
    @param pymodulename: the desired output module name
  '''
  def __init__(self, cleaned, pymodulename='ctypes_linux_generated'):
    if not os.access(cleaned, os.F_OK):
      raise IOError('The cleaned file %s does not exist'%(cleaned))
    self.cleaned = cleaned
    self.py = pymodulename
    self.xmlfile = '%s.%s'%(self.py, 'xml')
    self.pyfile = '%s.%s'%(self.py, 'py')
    self.gccxml = 'gccxml'
    self.xml2py = 'xml2py'
  
  def makeXml(self):
    cmd_line = [ self.gccxml, self.cleaned, '-fxml=%s'%(self.xmlfile), '-fextended-identifiers', '-fpreprocessed']
    p = subprocess.Popen(cmd_line, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    p.wait()
    build = p.stderr.read().strip()
    #print 'makexml', build
    if len(build) == 0:
      log.info( "GENERATED XML %s"%(self.xmlfile))
    else:
      log.error('Please clean %s\n%s'%(self.cleaned, build))
    return len(build)

  def makePy(self):
    if not os.access(self.xmlfile, os.F_OK):
      log.error('The XML file %s has not been generated'%(self.xml))
      return -1
    cmd_line = [self.xml2py, self.xmlfile, '-o', self.pyfile, '-k', 'd', '-k', 'e', '-k', 's', '-k', 't']
    p = subprocess.Popen(cmd_line, stdin=None, stdout=subprocess.PIPE, stderr=subprocess.PIPE, close_fds=True)
    p.wait()
    build = p.stderr.read().strip()
    if len(build) == 0:
      log.warning('\n%s'%build)
    log.info( "GENERATED PY module %s"%(self.pyfile))
    return 0

  def run(self):
    ret = self.makeXml()
    if  ret > 0:
      return ret
    ret = self.makePy()
    if ret > 0:
      return ret
    log.info('GENERATION Done. Enjoy %s'%(self.pyfile))
    return 0

def gen( cleaned, modulename):
  p = Generator(cleaned, modulename)
  return p.run()


def make(sourcefile, modulename):
  import cleaner, preprocess
  if not os.access(sourcefile, os.F_OK):
    raise IOError(sourcefile)
  #sourcefile
  basename = os.path.basename(sourcefile)
  preprocessed = "%s.c"%(modulename)
  cleaned = "%s_clean.c"%(modulename)
  #xml = "%s.xml"%(modulename)
  pyfinal = "%s.py"%(modulename)
  if not os.access(pyfinal, os.F_OK):
    if not os.access(cleaned, os.F_OK):
      if not os.access(preprocessed, os.F_OK):
        # preprocess the file
        if preprocess.process(sourcefile, preprocessed) > 0:
          return
      log.info('PREPROCESS - OK')
      # clean it
      if cleaner.clean(preprocessed, cleaned) > 0:
        return
    log.info('CLEAN - OK')
    # generate yfinal
    if gen(cleaned, modulename) > 0:
      return
  log.info('PYFINAL - OK')
  __import__(modulename)
  import inspect
  nbClass = len(inspect.getmembers(sys.modules[modulename], inspect.isclass))
  nbMembers = len(inspect.getmembers(sys.modules[modulename], inspect.isclass))
  log.info("module %s has %d members for %d class"%(modulename, nbMembers, nbClass))

logging.basicConfig(level=logging.INFO)

#generate.gen('ctypes_linux_generated_clean.c','ctypes_linux_generated')

# generate.make('ctypes_linux.c','ctypes_linux_generated')
make('ctypes_openssl.c','ctypes_openssl_generated')




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import os,logging,sys, argparse

import ctypes
import ctypes_openssl,ctypes_openssh

from ctypes import * # TODO delete
from ptrace.ctypes_libc import libc
from haystack.abouchet import StructFinder
from haystack.memory_mapper import MemoryMapper
from output import FileWriter

# linux only
from ptrace.debugger.debugger import PtraceDebugger
from ptrace.debugger.memory_mapping import readProcessMappings

log=logging.getLogger('openssl')

from ctypes.util import find_library
_libssl = find_library('ssl')
_libssl = ctypes.cdll.LoadLibrary(_libssl)


class RSAFileWriter(FileWriter):
  def __init__(self,folder='outputs'):
    FileWriter.__init__(self,'id_rsa','key',folder)
  def writeToFile(self,instance):
    prefix=self.prefix
    '''
    PEM_write_RSAPrivateKey(f, rsa_p, None, None, 0, None, None)
    PEM_write_RSAPrivateKey(fp,x, [enc,kstr,klen,cb,u]) 
     -> PEM_ASN1_write((int (*)())i2d_RSAPrivateKey,PEM_STRING_RSA,fp, (char *)x, [enc,kstr,klen,cb,u])
    int PEM_ASN1_write(i2d_of_void *i2d, const char *name, FILE *fp, char *x, [const EVP_CIPHER *, unsigned char *kstr,int , pem_password_cb *, void *])
     -> PEM_ASN1_write_bio(i2d, name, b, x  [,enc,kstr,klen,callback,u] );
    int PEM_ASN1_write_bio(i2d_of_void *i2d, const char *name, BIO *bp, char *x ,  [..]
     -> i2d_RSAPrivateKey( sur x )
      -> ASN1_item_i2d_bio(ASN1_ITEM_rptr(RSAPrivateKey), bp, rsa);
       -> i=BIO_write(out,&(b[j]),n);
     -> i=PEM_write_bio(bp,name,buf,data,i);
     
    en gros, c'est ctypes_openssl.RSA().writeASN1(file)
    '''
    filename=self.get_valid_filename()
    f=libc.fopen(filename,"w")  
    ret=_libssl.PEM_write_RSAPrivateKey(f, ctypes.byref(instance), None, None, 0, None, None)
    libc.fclose(f)
    if ret < 1:
      log.error("Error saving key to file %s"% filename)
      return False
    log.info ("[X] Key saved to file %s"%filename)
    return True

class DSAFileWriter(FileWriter):
  def __init__(self,folder='outputs'):
    FileWriter.__init__(self,'id_dsa','key',folder)
  def writeToFile(self,instance):
    prefix=self.prefix
    filename=self.get_valid_filename()
    f=libc.fopen(filename,"w")
    ret=_libssl.PEM_write_DSAPrivateKey(f, ctypes.byref(instance), None, None, 0, None, None)
    if ret < 1:
      log.error("Error saving key to file %s"% filename)
      return False
    log.info ("[X] Key saved to file %s"%filename)
    return True
  

class OpenSSLStructFinder(StructFinder):
  ''' Must not fork to ptrace. We need the real ctypes structs '''
  # interesting structs
  rsaw=RSAFileWriter()
  dsaw=DSAFileWriter()  
  def __init__(self,pid):
    StructFinder.__init__(self,pid)
    self.OPENSSL_STRUCTS={     # name, ( struct, callback)
      'RSA': (ctypes_openssl.RSA, self.rsaw.writeToFile ),
      'DSA': (ctypes_openssl.DSA, self.dsaw.writeToFile )
      }
  def findAndSave(self, maxNum=1, fullScan=False, nommap=False):
    log.debug('look for RSA keys')
    outs=self.find_struct(ctypes_openssl.RSA, maxNum=maxNum, fullScan=fullScan )
    for rsa,addr in outs:
      self.save(rsa)    
    log.debug('look for DSA keys')
    outs=self.find_struct(ctypes_openssl.DSA, maxNum=maxNum, fullScan=fullScan )
    for dsa,addr in outs:
      self.save(dsa)    
    return
  #'BIGNUM':   'RSA': (ctypes_openssl.BIGNUM, )
  def save(self,instance):
    if type(instance) == ctypes_openssl.RSA:
      self.rsaw.writeToFile(instance)
    elif type(instance) == ctypes_openssl.DSA:
      self.dsaw.writeToFile(instance)
    else:
      log.error('I dont know how to save that : %s'%(instance))

def usage(txt):
  log.error("Usage : %s <pid> [offset] # find SSL Structs in process"% txt)
  sys.exit(-1)


def argparser():
  parser = argparse.ArgumentParser(prog='openssl.py', description='Capture of RSA and DSA keys.')
  parser.add_argument('pid', type=int, help='Target PID')
  parser.add_argument('--nommap', dest='mmap', action='store_const', const=False, default=True, help='disable mmap()-ing')
  parser.add_argument('--debug', dest='debug', action='store_const', const=True, help='setLevel to DEBUG')
  parser.set_defaults(func=search)
  return parser


def search(args):
  if args.debug:
    logging.basicConfig(level=logging.DEBUG)    
  log.info("Target has pid %d"%args.pid)
  mappings = MemoryMapper(args).getMappings()
  finder = OpenSSLStructFinder(mappings )
  outs=finder.findAndSave()
  return



def main(argv):
  logging.basicConfig(level=logging.INFO)

  # use optarg on v, a and to
  parser = argparser()
  opts = parser.parse_args(argv)
  try:
    opts.func(opts)
  except ImportError,e:
    log.error('Struct type does not exists.')
    print e

        
  log.info("done for pid %d"%opts.pid)

  return -1


if __name__ == "__main__":
  main(sys.argv[1:])




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2011 Loic Jaquemet loic.jaquemet+python@gmail.com
#

__author__ = "Loic Jaquemet loic.jaquemet+python@gmail.com"

import os,logging,sys,time,io,select,socket
import threading
from threading import Thread

from paramiko_packet import NeedRekeyException
from paramiko.ssh_exception import SSHException
from paramiko.common import *

from socket_scapy import MissingDataException, MISSING_DATA_MESSAGE

from ctypes_openssh import AES_BLOCK_SIZE

log=logging.getLogger('output')

MAX_KEYS=255


class FileWriter:
  def __init__(self,prefix,suffix,folder):
    self.prefix=prefix
    self.suffix=suffix
    self.folder=folder
  def get_valid_filename(self):
    filename_FMT="%s-%d.%s"
    for i in xrange(1,MAX_KEYS):
      filename=filename_FMT%(self.prefix,i,self.suffix)
      afilename=os.path.normpath(os.path.sep.join([self.folder,filename]))
      if not os.access(afilename,os.F_OK):
        return afilename
    #
    log.error("Too many file keys extracted in %s directory"%(self.folder))
    return None    
  def writeToFile(self,instance):
    raise NotImplementedError

class SSHStreamToFile():
  ''' Pipes the data from a (ssh) socket into a different file for each packet type. 
    supposedly, this would demux channels into files.
    We still need to differenciate at higher level, with two SSHStreamFile
     between upload and download.
  '''
  BUFSIZE=4096
  def __init__(self, packetizer, ctx, basename, folder='outputs', fmt="%Y%m%d-%H%M%S"):
    self.packetizer = packetizer
    #self.refresher = refresher
    self.datename = "%s"%time.strftime(fmt,time.gmtime())
    self.fname=os.path.sep.join([folder,basename])
    self.outs=dict()
    self.engine=ctx['engine']
    self.socket=ctx['socket']
    ##
    self.lastMessage=None
    self.decrypt_errors=0
    return

  def process(self):
    data=self._in.read(self.BUFSIZE)
    self._out.write(data2)
  
  def _outputStream(self, channel):
    name="%s.%s.%d"%(self.fname, self.datename, channel )
    if name in self.outs:
      return self.outs[name]
    else:
      self.outs[name] = io.FileIO(name , 'w' )
      log.info("New Output Filename is %s"%(name))
    log.debug("Output Filename is %s"%(name))
    return self.outs[name]

  def process(self):
    try:
      m = self._process()
      if self.decrypt_errors:
        log.info("we read %d blocks/%d bytes and couldn't make sense out of it"%(self.decrypt_errors, self.decrypt_errors*16 ))
        log.info("But we made it : to %s"%(str(m) ) )
        self.decrypt_errors = 0
    except SSHException,e:  # only size errror... no sense. should be only one exception. 
      self.decrypt_errors+=1
      log.debug('SSH exception catched on %s - %s - will try to find next good Message'%(self.fname,e))
      return


  def _process(self):
    ''' m can be rewind()-ed , __str__ ()-ed or others...
  '''
    _expected_packet = tuple()
    try:
      ptype, m = self.packetizer.read_message()
      self.lastMessage=m
      #log.error("now  message was (%d) : %s"%(len(str(m)),repr(str(m))) )
      #self.lastCounter=self.engine.getCounter()
      if ptype != 94:
        log.warning("===================== ptype:%d len:%d "%(ptype, len(str(m)) ) )
    except NeedRekeyException,e:
      log.warning('=============================== Please refresh keys for rekey')
      return e
    except OverflowError,e:
      log.warning('SSH exception catched/bad packet size on %s'%(self.fname))
      #self.refresher.refresh()
      return e
    except MissingDataException, e:
      log.warning('=============================== Missing data. Please refresh keys for rekey')
      return e
    if ptype == MSG_IGNORE:
      log.warning('================================== MSG_IGNORE')
      return 'MSG_IGNORE'
    elif ptype == MSG_DISCONNECT:
      log.info( "==================================== DISCONNECT MESSAGE")
      log.info( m)
      self.packetizer.close()
      return 'MSG_DISCONNECT'
    elif ptype == MSG_DEBUG:
      always_display = m.get_boolean()
      msg = m.get_string()
      lang = m.get_string()
      log.warning('Debug msg: ' + util.safe_string(msg))
      return 'MSG_DEBUG'
    if len(_expected_packet) > 0:
      if ptype not in _expected_packet:
        raise SSHException('Expecting packet from %r, got %d' % (_expected_packet, ptype))
      _expected_packet = tuple()
      if (ptype >= 30) and (ptype <= 39):
        log.info("KEX Message, we need to rekey")
        return 'KEX'
    #
    out=self._outputStream(ptype)
    ret=out.write( str(m) )
    out.flush() # beuahhh
    log.debug("%d bytes written for channel %d"%(ret, ptype))
    return m

  
class Supervisor(threading.Thread):
  def __init__(self ):
    self.stopSwitch=threading.Event()
    self.readables=dict()
    self.selectables=set()
    self._readables=dict()
    self._selectables=set()
    self.lock=threading.Lock()
    self.todo=False
    return
  
  def add(self, socket, handler):
    '''
      @param soket: the socket to select() onto  
      @param handler: the callable to run when data arrives.
    '''
    self.lock.acquire()
    self.readables[socket] = handler
    self.selectables.add(socket)
    self.todo=True
    self.lock.release()
    return 

  def _syncme(self):
    self.lock.acquire()
    self._readables = dict(self.readables)
    self._selectables = set(self.selectables)
    self.todo=False
    self.lock.release()
    return 

  def run(self):
     # thread inbound reads and writes  
    while not self.stopSwitch.isSet():
      # check
      if self.todo:
        self._syncme
      r,w,o=select.select(self.selectables,[],[],2)
      if len(r) == 0:
        log.debug("select waked up without anything to read... going back to select()")
        continue
      # read them and write them
      for soket in r:
        self.readables[soket]()
        log.debug("read and write done for %s"%(soket))
      #loop
    log.info('Supervisor finished running') 
    return


